#
# Copyright (C) 2018-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Convert a HEALPix FITS file to multi-resolution UNIQ indexing from the more
common IMPLICIT indexing.
"""

from argparse import FileType

from . import ArgumentParser


def parser():
    parser = ArgumentParser()
    parser.add_argument(
        "input", metavar="INPUT.fits", type=FileType("rb"), help="Input FITS file"
    )
    parser.add_argument(
        "output",
        metavar="OUTPUT.fits[.gz]",
        type=FileType("wb"),
        help="Output FITS file",
    )
    return parser


def main(args=None):
    with parser().parse_args(args) as args:
        import warnings

        from astropy.io import fits

        from ..io import read_sky_map, write_sky_map

        hdus = fits.open(args.input)
        ordering = hdus[1].header["ORDERING"]
        expected_orderings = {"NESTED", "RING"}
        if ordering not in expected_orderings:
            msg = "Expected the FITS file {} to have ordering {}, but it is {}"
            warnings.warn(
                msg.format(args.input.name, " or ".join(expected_orderings), ordering)
            )
        table = read_sky_map(hdus, moc=True)
        write_sky_map(args.output.name, table)



#
# Copyright (C) 2019-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Make an observability chart for a LIGO/Virgo/KAGRA probability sky map."""

from argparse import FileType

import numpy as np

from . import ArgumentParser, HelpChoicesAction
from .matplotlib import get_figure_parser


def parser():
    from astropy.coordinates import EarthLocation

    site_names = EarthLocation.get_site_names()
    parser = ArgumentParser(parents=[get_figure_parser()])
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Print airmass table to stdout"
    )
    parser.add_argument(
        "input",
        metavar="INPUT.fits[.gz]",
        type=FileType("rb"),
        default="-",
        nargs="?",
        help="Input FITS file",
    )
    parser.add_argument("--time", help="UTC time")
    parser.add_argument(
        "--max-airmass", default=2.5, type=float, help="Maximum airmass"
    )
    parser.add_argument(
        "--twilight",
        default="astronomical",
        choices=("astronomical", "nautical", "civil"),
        help="Twilight definition: astronomical (-18 degrees), "
        "nautical (-12 degrees), or civil (-6 degrees)",
    )
    parser.add_argument(
        "--site",
        nargs="*",
        default=[],
        metavar="SITE",
        choices=site_names,
        help="Observatory site",
    )
    parser.add_argument("--help-site", action=HelpChoicesAction, choices=site_names)
    parser.add_argument("--site-name", nargs="*", default=[], help="Observatory name.")
    parser.add_argument(
        "--site-longitude",
        nargs="*",
        default=[],
        metavar="DEG",
        type=float,
        help="Observatory longitude on the WGS84 ellipsoid.",
    )
    parser.add_argument(
        "--site-latitude",
        nargs="*",
        default=[],
        metavar="DEG",
        type=float,
        help="Observatory latitude on the WGS84 ellipsoid.",
    )
    parser.add_argument(
        "--site-height",
        nargs="*",
        default=[],
        metavar="METERS",
        type=float,
        help="Observatory height from the WGS84 ellipsoid.",
    )
    return parser


def condition_secz(x):
    """Condition secz airmass formula: values <= 0 are below the horizon,
    which we map to infinite airmass.
    """
    return np.where(x <= 0, np.inf, x)


def main(args=None):
    p = parser()
    with p.parse_args(args) as opts:
        # Late imports
        from astroplan import (
            AirmassConstraint,
            AtNightConstraint,
            Observer,
            is_event_observable,
        )
        from astropy import units as u
        from astropy.coordinates import EarthLocation, SkyCoord
        from astropy.time import Time
        from matplotlib import dates
        from matplotlib import pyplot as plt
        from tqdm import tqdm

        from .. import (
            moc,
            plot,  # noqa: F401
        )
        from ..io import fits

        names = ("name", "longitude", "latitude", "height")
        length0, *lengths = (
            len(getattr(opts, "site_{}".format(name))) for name in names
        )
        if not all(length0 == length for length in lengths):
            p.error(
                "these options require equal numbers of arguments: {}".format(
                    ", ".join("--site-{}".format(name) for name in names)
                )
            )

        observers = [Observer.at_site(site) for site in opts.site]
        for name, lon, lat, height in zip(
            opts.site_name, opts.site_longitude, opts.site_latitude, opts.site_height
        ):
            location = EarthLocation(
                lon=lon * u.deg, lat=lat * u.deg, height=(height or 0) * u.m
            )
            observers.append(Observer(location, name=name))
        observers = list(reversed(observers))

        m = fits.read_sky_map(opts.input.name, moc=True)

        t0 = Time(opts.time) if opts.time is not None else Time.now()
        times = t0 + np.linspace(0, 1) * u.day

        theta, phi = moc.uniq2ang(m["UNIQ"])
        coords = SkyCoord(phi, 0.5 * np.pi - theta, unit=u.rad)
        prob = np.asarray(moc.uniq2pixarea(m["UNIQ"]) * m["PROBDENSITY"])

        constraints = [
            getattr(AtNightConstraint, "twilight_{}".format(opts.twilight))(),
            AirmassConstraint(opts.max_airmass),
        ]

        width, _ = plt.rcParams["figure.figsize"]
        height = (len(observers) + 1) / 16 * width
        ax = plt.figure(figsize=(width, height)).add_subplot()
        locator = dates.AutoDateLocator()
        formatter = dates.DateFormatter("%H:%M")
        ax.set_xlim([times[0].plot_date, times[-1].plot_date])
        ax.xaxis.set_major_formatter(formatter)
        ax.xaxis.set_major_locator(locator)
        ax.set_xlabel("Time from {0} [UTC]".format(min(times).datetime.date()))
        plt.setp(ax.get_xticklabels(), rotation=30, ha="right")
        ax.set_yticks(np.arange(len(observers)))
        ax.set_yticklabels([observer.name for observer in observers])
        ax.yaxis.set_tick_params(left=False)
        ax.grid(axis="x")
        ax.spines["bottom"].set_visible(False)
        ax.spines["top"].set_visible(False)

        for i, observer in enumerate(tqdm(observers)):
            observable = 100 * np.dot(
                prob, is_event_observable(constraints, observer, coords, times)
            )
            ax.contourf(
                times.plot_date,
                [i - 0.4, i + 0.4],
                np.tile(observable, (2, 1)),
                levels=np.arange(10, 110, 10),
                cmap=plt.get_cmap().reversed(),
            )

        plt.tight_layout()

        # Show or save output.
        opts.output()



#
# Copyright (C) 2018-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Convert a HEALPix FITS file from multi-resolution UNIQ indexing to the more
common IMPLICIT indexing.
"""

from argparse import FileType

from . import ArgumentParser


def parser():
    parser = ArgumentParser()
    parser.add_argument(
        "input", metavar="INPUT.fits[.gz]", type=FileType("rb"), help="Input FITS file"
    )
    parser.add_argument(
        "output", metavar="OUTPUT.fits", type=FileType("wb"), help="Output FITS file"
    )
    parser.add_argument("--nside", type=int, help="Output HEALPix resolution")
    return parser


def main(args=None):
    with parser().parse_args(args) as args:
        import logging
        import warnings

        import astropy_healpix as ah
        from astropy.io import fits

        from ..bayestar import rasterize
        from ..io import read_sky_map, write_sky_map

        log = logging.getLogger()

        if args.nside is None:
            order = None
        else:
            order = ah.nside_to_level(args.nside)

        log.info("reading FITS file %s", args.input.name)
        hdus = fits.open(args.input)
        ordering = hdus[1].header["ORDERING"]
        expected_ordering = "NUNIQ"
        if ordering != expected_ordering:
            msg = "Expected the FITS file {} to have ordering {}, but it is {}"
            warnings.warn(msg.format(args.input.name, expected_ordering, ordering))
        log.debug("converting original FITS file to Astropy table")
        table = read_sky_map(hdus, moc=True)
        log.debug("flattening HEALPix tree")
        table = rasterize(table, order=order)
        log.info("writing FITS file %s", args.output.name)
        write_sky_map(args.output.name, table, nest=True)
        log.debug("done")



import errno
import gzip
import os
import re
import subprocess
from contextlib import ExitStack
from pathlib import Path

import h5py
import numpy as np
import pytest
from igwn_ligolw.utils import load_filename

from ...util import sqlite
from .. import events
from ..events.ligolw import ContentHandler

DATA_PATH = os.path.join(os.path.dirname(__file__), "data")


@pytest.fixture
def mock_gracedb(monkeypatch):
    stack = ExitStack()

    class MockGraceDb(ExitStack):
        def files(self, graceid, filename):
            path = os.path.join(DATA_PATH, "{}_{}".format(graceid, filename))
            try:
                return stack.enter_context(open(path, "rb"))
            except IOError as e:
                if e.errno != errno.ENOENT:
                    raise
                return stack.enter_context(gzip.GzipFile(path + ".gz", "rb"))

    with stack:
        monkeypatch.setattr("ligo.gracedb.rest.GraceDb", MockGraceDb)
        yield


def raises(expected_exception, msg):
    return pytest.raises(expected_exception, match="^" + re.escape(msg) + "$")


def ligolw_assertions(source):
    """Test common assertions for test_ligolw and test_sqlite."""
    assert len(source) == 250
    event = source[821759]
    assert len(event.singles) == 2
    assert event.singles[0].snr_series is event.singles[1].snr_series is None
    assert event.singles[0].detector == "H1"
    assert event.singles[0].snr == 8.5362396
    assert event.singles[0].phase == -0.81192881
    assert event.singles[0].time == event.singles[0].zerolag_time == 967328914.866107842
    psd = event.singles[0].psd
    assert psd.f0 == 0.0
    assert psd.deltaF == 0.125
    assert event.singles[1].detector == "L1"
    assert event.singles[1].snr == 10.36818
    assert event.singles[1].phase == 1.9740163
    assert event.singles[1].time == event.singles[1].zerolag_time == 967328914.866513726
    psd = event.singles[0].psd
    assert psd.f0 == 0.0
    assert psd.deltaF == 0.125
    assert event.template_args == {
        "mass1": 1.56687,
        "mass2": 1.474779,
        "spin1x": 0.0,
        "spin1y": 0.0,
        "spin1z": 0.0,
        "spin2x": 0.0,
        "spin2y": 0.0,
        "spin2z": 0.0,
        "f_final": 2047.0,
    }


def test_unknown():
    with pytest.raises(IOError, match="Unknown file format"):
        events.open(__file__)


def test_ligolw():
    """Test reading events from LIGO-LW XML files."""
    source = events.open(os.path.join(DATA_PATH, "2016_subset.xml.gz"))
    ligolw_assertions(source)

    assert str(source).startswith("<LigoLWEventSource({288172: <LigoLWEvent(")
    assert repr(source).startswith("LigoLWEventSource({288172: <LigoLWEvent(")
    event, *_ = source.values()
    assert str(event).startswith(
        "<LigoLWEvent(singles=(<LigoLWSingleEvent(detector='H1', snr=12."
    )
    assert repr(event).startswith(
        "<LigoLWEvent(singles=(<LigoLWSingleEvent(detector='H1', snr=12."
    )
    single_event, *_ = event.singles
    assert (
        str(single_event)
        == "<LigoLWSingleEvent(detector='H1', snr=12.035994, phase=-1.0371021, time=970976257.4808338)>"
    )
    assert (
        repr(single_event)
        == "<LigoLWSingleEvent(detector='H1', snr=12.035994, phase=-1.0371021, time=970976257.4808338)>"
    )


def test_ligolw_document():
    """Test reading events from LIGO-LW XML document."""
    xmldoc = load_filename(
        os.path.join(DATA_PATH, "2016_subset.xml.gz"), contenthandler=ContentHandler
    )
    source = events.open(xmldoc, fallbackpath=DATA_PATH)
    ligolw_assertions(source)


def test_ligolw_pathlib():
    """Test reading file provided with Pathlib instead of string"""
    source = events.open(Path(DATA_PATH, "2016_subset.xml.gz"))
    ligolw_assertions(source)


def test_sqlite(tmpdir):
    """Test reading events from SQLite files."""
    # Convert the test data to SQLite format in the temporary directory.
    xmlfilename = os.path.join(DATA_PATH, "2016_subset.xml.gz")
    dbfilename = str(tmpdir / "2016_subset.sqlite")
    subprocess.check_call(["igwn_ligolw_sqlite", "-p", xmlfilename, "-d", dbfilename])

    # Symbolicly link the PSD directory into the temporary directory.
    os.symlink(
        os.path.join(DATA_PATH, "gstlal_reference_psd"),
        str(tmpdir / "gstlal_reference_psd"),
    )

    source = events.open(dbfilename)
    ligolw_assertions(source)

    with open(dbfilename, "rb") as f:
        source = events.open(f)
        ligolw_assertions(source)

    with sqlite.open(dbfilename, "r") as db:
        source = events.open(db)
        ligolw_assertions(source)


def test_gracedb(mock_gracedb):
    """Test reading events from GraceDB records."""
    source = events.gracedb.open(["G211117", "G197392"])
    assert len(source) == 2
    for i, (event_id, event) in enumerate(source.items()):
        if i == 0:
            assert event_id == "G211117"
            assert event.singles[0].snr_series is event.singles[1].snr_series is None
            assert event.singles[0].detector == "H1"
            assert event.singles[0].snr == 9.0802174
            assert event.singles[0].phase == -0.13969257
            assert (
                event.singles[0].time
                == event.singles[0].zerolag_time
                == 1135136350.647757924
            )
            psd = event.singles[0].psd
            assert psd.f0 == 0.0
            assert psd.deltaF == 0.125
            assert event.singles[1].detector == "L1"
            assert event.singles[1].snr == 7.3947201
            assert event.singles[1].phase == -2.7356486
            assert (
                event.singles[1].time
                == event.singles[1].zerolag_time
                == 1135136350.646883043
            )
            psd = event.singles[1].psd
            assert psd.f0 == 0.0
            assert psd.deltaF == 0.125
            assert event.template_args == {
                "mass1": 19.924686,
                "mass2": 6.4254546,
                "spin1x": 0.0,
                "spin1y": 0.0,
                "spin1z": 0.33962944,
                "spin2x": 0.0,
                "spin2y": 0.0,
                "spin2z": -0.1238557,
                "f_final": 1024.0,
            }
        elif i == 1:
            assert event_id == "G197392"
            assert event.singles[0].snr_series is event.singles[1].snr_series is None
            assert event.singles[0].detector == "H1"
            assert event.singles[0].snr == 6.9068823
            assert event.singles[0].phase == 1.8298783
            assert (
                event.singles[0].time
                == event.singles[0].zerolag_time
                == 1128678900.444335938
            )
            psd = event.singles[0].psd
            assert psd.f0 == 30.0
            assert psd.deltaF == 0.125
            assert event.singles[1].detector == "L1"
            assert event.singles[1].snr == 6.8389997
            assert event.singles[1].phase == -1.0297496
            assert (
                event.singles[1].time
                == event.singles[1].zerolag_time
                == 1128678900.445068359
            )
            psd = event.singles[1].psd
            assert psd.f0 == 30.0
            assert psd.deltaF == 0.125
            assert event.template_args == {
                "mass1": 32.064007,
                "mass2": 14.607587,
                "spin1x": 0.0,
                "spin1y": 0.0,
                "spin1z": 0.34881824,
                "spin2x": 0.0,
                "spin2y": 0.0,
                "spin2z": -0.53029484,
                "f_final": 0.0,
            }


def test_detector_disabled(mock_gracedb):
    """Test reading from event sources with certain detectors disabled."""
    graceids = ("G211117", "G197392")
    base_source = events.gracedb.open(graceids)

    source = events.detector_disabled.open(base_source, ["H1"])
    assert len(source) == 2
    for graceid, (event_id, event) in zip(graceids, source.items()):
        assert event_id == graceid
        assert len(event.singles) == 1
        assert event.singles[0].detector == "L1"

    for event, base_event in zip(source.values(), base_source.values()):
        assert event.template_args == base_event.template_args

    # Now test that exceptions are raised when they are called for.
    expected_message = (
        "Disabling detectors {H1, L1} would have no effect on "
        "this event with detectors {H1 L1}"
    )
    nonraising_source = events.detector_disabled.open(
        base_source, ["H1, L1"], raises=False
    )
    raising_source = events.detector_disabled.open(base_source, ["H1, L1"])
    for event in nonraising_source.values():
        event.singles
    for event in raising_source.values():
        with raises(events.DetectorDisabledError, expected_message):
            event.singles

    # Now test that exceptions are raised when they are called for.
    expected_message = (
        "Disabling detectors {H1 L1 V1} would exclude all "
        "data for this event with detectors {H1 L1}"
    )
    nonraising_source = events.detector_disabled.open(
        base_source, ["H1", "L1", "V1"], raises=False
    )
    raising_source = events.detector_disabled.open(base_source, ["H1", "L1", "V1"])
    for event in nonraising_source.values():
        event.singles
    for event in raising_source.values():
        with raises(events.DetectorDisabledError, expected_message):
            event.singles


def test_hdf(tmpdir):
    """Test reading events from HDF5 files."""
    # Create test input files
    ifos = ["L1", "H1"]
    filenames = []
    filename = str(tmpdir / "coincs.hdf")
    filenames.append(filename)
    with h5py.File(filename, "w") as coinc_file:
        coinc_file.attrs["timeslide_interval"] = np.e
        coinc_group = coinc_file.create_group("foreground")
        coinc_group["template_id"] = np.arange(5)
        coinc_group["timeslide_id"] = np.arange(0, 50, 10)

        filename = str(tmpdir / "H1L1-BANK.hdf")
        filenames.append(filename)
        with h5py.File(filename, "w") as bank_file:
            bank_file.attrs["parameters"] = []

        for i, ifo in enumerate(ifos):
            coinc_file.attrs["detector_{}".format(i + 1)] = ifo
            coinc_group["trigger_id{}".format(i + 1)] = np.arange(5)

            filename = str(tmpdir / (ifo + "_triggers.hdf"))
            filenames.append(filename)
            with h5py.File(filename, "w") as trigger_file:
                trigger_group = trigger_file.create_group(ifo)
                trigger_group["snr"] = i + np.arange(5) * np.pi
                trigger_group["coa_phase"] = i + np.arange(5) * np.pi**2
                trigger_group["end_time"] = i + np.arange(5) * np.pi**3

            filename = str(tmpdir / (ifo + "_psds.hdf"))
            filenames.append(filename)
            with h5py.File(filename, "w") as psd_file:
                psd_file.attrs["low_frequency_cutoff"] = 10.0
                psd_file.attrs["dynamic_range_factor"] = np.e
                psd_group = psd_file.create_group(ifo)
                psd_group["start_time"] = (np.arange(5) - 0.5) * np.pi**3
                psd_group["end_time"] = (np.arange(5) + 0.5) * np.pi**3
                psd_group = psd_group.create_group("psds")
                for j in range(5):
                    psd_group[str(j)] = (
                        np.concatenate((np.zeros(5), np.arange(5) ** 2)) * np.e**2
                    )
                    psd_group[str(j)].attrs["delta_f"] = 2.0

    with pytest.raises(ValueError, match="You must provide exactly one coinc file."):
        source = events.open(*filenames[1:])

    with pytest.raises(
        ValueError, match="You must provide exactly one template bank file."
    ):
        source = events.open(*filenames[:1])

    with pytest.raises(ValueError, match="You must provide PSD files."):
        source = events.open(*filenames[:2])

    with pytest.raises(ValueError, match="You must provide trigger files."):
        source = events.open(*filenames[:2], filenames[3], filenames[5])

    # Test reading from filenames
    source = events.open(*filenames)
    assert len(source) == 5
    for coinc_id, coinc in source.items():
        for i, (ifo, single) in enumerate(zip(ifos, coinc.singles)):
            assert single.detector == ifo
            assert single.snr == i + coinc_id * np.pi
            assert single.phase == i + coinc_id * np.pi**2
            assert single.time == (
                single.zerolag_time + coinc_id * 10 * np.e * (-0.5 if i == 0 else +0.5)
            )
            assert single.zerolag_time == i + coinc_id * np.pi**3
            assert single.psd.f0 == 10.0
            assert single.psd.deltaF == 2.0
            assert np.all(single.psd.data.data == np.arange(5) ** 2)
        assert coinc.template_args == {}

    # Test reading from h5py.File instances
    events.open(*(h5py.File(filename, "r") for filename in filenames))

    # Test reading from file-like objects
    with ExitStack() as stack:
        events.open(
            *(stack.enter_context(open(filename, "rb")) for filename in filenames)
        )



#
# Copyright (C) 2019-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
List most likely constellations for a localization.

Just for fun, and for public outreach purposes.
"""

from argparse import FileType

from . import ArgumentParser


def parser():
    parser = ArgumentParser()
    parser.add_argument(
        "input",
        metavar="INPUT.fits[.gz]",
        type=FileType("rb"),
        default="-",
        nargs="?",
        help="Input FITS file",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="OUT.dat",
        type=FileType("w"),
        default="-",
        help="Name of output file",
    )
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        # Late imports
        import astropy_healpix as ah
        import healpy as hp
        import numpy as np
        from astropy import units as u
        from astropy.coordinates import SkyCoord
        from astropy.table import Table

        from ..io import fits

        prob, meta = fits.read_sky_map(opts.input.name, nest=None)
        npix = len(prob)
        nside = ah.npix_to_nside(npix)
        ipix = np.arange(npix)
        ra, dec = hp.pix2ang(nside, ipix, lonlat=True, nest=meta["nest"])
        coord = SkyCoord(ra * u.deg, dec * u.deg)
        table = Table(
            {"prob": prob, "constellation": coord.get_constellation()}, copy=False
        )
        table = table.group_by("constellation").groups.aggregate(np.sum)
        table.sort("prob")
        table.reverse()
        table.write(opts.output, format="ascii.tab")



# Copyright (C) 2017-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Read events from PyCBC-style HDF5 output."""

from itertools import groupby
from operator import itemgetter

import h5py
import lal
import numpy as np
from igwn_segments import segment, segmentlist

from .base import Event, EventSource, SingleEvent

__all__ = ("HDFEventSource",)


class _psd_segment(segment):  # noqa: N801
    def __new__(cls, psd, *args):
        return segment.__new__(cls, *args)

    def __init__(self, psd, *args):
        self.psd = psd


def _hdf_file(f):
    if isinstance(f, h5py.File):
        return f
    elif hasattr(f, "read") and hasattr(f, "name"):
        return h5py.File(f.name, "r")
    else:
        return h5py.File(f, "r")


def _classify_hdf_file(f, sample):
    if sample in f:
        return "coincs"
    for key, value in f.items():
        if isinstance(value, h5py.Group):
            if "psds" in value:
                return "psds"
            if "snr" in value and "coa_phase" in value and "end_time" in value:
                return "triggers"
    if "parameters" in f.attrs:
        return "bank"
    raise ValueError("Unrecognized PyCBC file type")


class HDFEventSource(EventSource):
    """Read events from PyCBC-style HDF5 files.

    Parameters
    ----------
    *files : list of str, file-like object, or `h5py.File` objects
        The PyCBC coinc, bank, psds, and triggers files, in any order.

    Returns
    -------
    `~ligo.skymap.io.events.EventSource`

    """

    def __init__(self, *files, **kwargs):
        sample = kwargs.get("sample", "foreground")

        # Open the files and split them into coinc files, bank files, psds,
        # and triggers.
        key = itemgetter(0)
        files = [_hdf_file(f) for f in files]
        files = sorted([(_classify_hdf_file(f, sample), f) for f in files], key=key)
        files = {key: list(v[1] for v in value) for key, value in groupby(files, key)}

        try:
            (coinc_file,) = files["coincs"]
        except (KeyError, ValueError):
            raise ValueError("You must provide exactly one coinc file.")
        try:
            (bank_file,) = files["bank"]
        except (KeyError, ValueError):
            raise ValueError("You must provide exactly one template bank file.")
        try:
            psd_files = files["psds"]
        except KeyError:
            raise ValueError("You must provide PSD files.")
        try:
            trigger_files = files["triggers"]
        except KeyError:
            raise ValueError("You must provide trigger files.")

        self._bank = bank_file

        key_prefix = "detector_"
        detector_nums, self._ifos = zip(
            *sorted(
                (int(key[len(key_prefix) :]), value)
                for key, value in coinc_file.attrs.items()
                if key.startswith(key_prefix)
            )
        )

        coinc_group = coinc_file[sample]
        self._timeslide_interval = coinc_file.attrs.get("timeslide_interval", 0)
        self._template_ids = coinc_group["template_id"]
        self._timeslide_ids = coinc_group.get("timeslide_id", np.zeros(len(self)))
        self._trigger_ids = [
            coinc_group["trigger_id{}".format(detector_num)]
            for detector_num in detector_nums
        ]

        triggers = {}
        for f in trigger_files:
            ((ifo, group),) = f.items()
            triggers[ifo] = [group["snr"], group["coa_phase"], group["end_time"]]
        self._triggers = tuple(triggers[ifo] for ifo in self._ifos)

        psdseglistdict = {}
        for psd_file in psd_files:
            ((ifo, group),) = psd_file.items()
            psd = [group["psds"][str(i)] for i in range(len(group["psds"]))]
            psdseglistdict[ifo] = segmentlist(
                _psd_segment(*segargs)
                for segargs in zip(psd, group["start_time"], group["end_time"])
            )
        self._psds = [psdseglistdict[ifo] for ifo in self._ifos]

    def __getitem__(self, id):
        return HDFEvent(self, id)

    def __iter__(self):
        return iter(range(len(self)))

    def __len__(self):
        return len(self._template_ids)


class HDFEvent(Event):
    def __init__(self, source, id):
        self._source = source
        self._id = id

    @property
    def singles(self):
        return tuple(
            HDFSingleEvent(
                ifo,
                self._id,
                i,
                trigger_ids[self._id],
                self._source._timeslide_interval,
                triggers,
                self._source._timeslide_ids,
                psds,
            )
            for i, (ifo, trigger_ids, triggers, psds) in enumerate(
                zip(
                    self._source._ifos,
                    self._source._trigger_ids,
                    self._source._triggers,
                    self._source._psds,
                )
            )
        )

    @property
    def template_args(self):
        bank = self._source._bank
        bank_id = self._source._template_ids[self._id]
        return {key: value[bank_id] for key, value in bank.items()}


class HDFSingleEvent(SingleEvent):
    def __init__(
        self,
        detector,
        _coinc_id,
        _detector_num,
        _trigger_id,
        _timeslide_interval,
        _triggers,
        _timeslide_ids,
        _psds,
    ):
        self._detector = detector
        self._coinc_id = _coinc_id
        self._detector_num = _detector_num
        self._trigger_id = _trigger_id
        self._timeslide_interval = _timeslide_interval
        self._triggers = _triggers
        self._timeslide_ids = _timeslide_ids
        self._psds = _psds

    @property
    def detector(self):
        return self._detector

    @property
    def snr(self):
        return self._triggers[0][self._trigger_id]

    @property
    def phase(self):
        return self._triggers[1][self._trigger_id]

    @property
    def time(self):
        value = self.zerolag_time

        # PyCBC does not specify which detector is time-shifted in time slides.
        # Since PyCBC's coincidence format currently supports only two
        # detectors, we arbitrarily apply half of the time slide to each
        # detector.
        shift = self._timeslide_ids[self._coinc_id] * self._timeslide_interval
        if self._detector_num == 0:
            value -= 0.5 * shift
        elif self._detector_num == 1:
            value += 0.5 * shift
        else:
            raise AssertionError("This line should not be reached")
        return value

    @property
    def zerolag_time(self):
        return self._triggers[2][self._trigger_id]

    @property
    def psd(self):
        try:
            psd = self._psds[self._psds.find(self.zerolag_time)].psd
        except ValueError:
            raise ValueError(
                "No PSD found for detector {} at zero-lag GPS time {}".format(
                    self.detector, self.zerolag_time
                )
            )

        dyn_range_fac = psd.file.attrs["dynamic_range_factor"]
        flow = psd.file.attrs["low_frequency_cutoff"]
        df = psd.attrs["delta_f"]
        kmin = int(flow / df)

        fseries = lal.CreateREAL8FrequencySeries(
            "psd", 0, kmin * df, df, lal.DimensionlessUnit, len(psd) - kmin
        )
        fseries.data.data = psd[kmin:] / np.square(dyn_range_fac)
        return fseries


open = HDFEventSource



#
# Copyright (C) 2013-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Markov-Chain Monte Carlo sky localization."""

from argparse import FileType

from . import (
    ArgumentParser,
    get_posterior_parser,
    get_random_parser,
    get_waveform_parser,
)


def parser():
    parser = ArgumentParser(
        parents=[get_waveform_parser(), get_posterior_parser(), get_random_parser()]
    )
    parser.add_argument(
        "input",
        metavar="INPUT.{hdf,xml,xml.gz,sqlite}",
        default="-",
        nargs="+",
        type=FileType("rb"),
        help="Input LIGO-LW XML file, SQLite file, or PyCBC HDF5 files. "
        "For PyCBC, you must supply the coincidence file "
        '(e.g. "H1L1-HDFINJFIND.hdf" or "H1L1-STATMAP.hdf"), '
        "the template bank file (e.g. H1L1-BANK2HDF.hdf), "
        "the single-detector merged PSD files "
        '(e.g. "H1-MERGE_PSDS.hdf" and "L1-MERGE_PSDS.hdf"), '
        "and the single-detector merged trigger files "
        '(e.g. "H1-HDF_TRIGGER_MERGE.hdf" and '
        '"L1-HDF_TRIGGER_MERGE.hdf"), '
        "in any order.",
    )
    parser.add_argument(
        "--pycbc-sample", default="foreground", help="(PyCBC only) sample population"
    )
    parser.add_argument(
        "--coinc-event-id",
        type=int,
        nargs="*",
        help="run on only these specified events",
    )
    parser.add_argument("--output", "-o", default=".", help="output directory")
    parser.add_argument(
        "--condor-submit",
        action="store_true",
        help="submit to Condor instead of running locally",
    )

    group = parser.add_argument_group(
        "fixed parameter options", "Options to hold certain parameters constant"
    )
    group.add_argument("--ra", type=float, metavar="DEG", help="Right ascension")
    group.add_argument("--dec", type=float, metavar="DEG", help="Declination")
    group.add_argument(
        "--distance", type=float, metavar="Mpc", help="Luminosity distance"
    )

    return parser


def identity(x):
    return x


def main(args=None):
    with parser().parse_args(args) as opts:
        import logging

        log = logging.getLogger("BAYESTAR")

        # BAYESTAR imports.
        import os
        import subprocess
        import sys

        import lal
        import numpy as np

        # Other imports.
        from astropy.table import Table

        from ..bayestar import condition, condition_prior, ez_emcee, log_post
        from ..io import events, hdf5

        # Read coinc file.
        log.info("%s:reading input files", ",".join(file.name for file in opts.input))
        event_source = events.open(*opts.input, sample=opts.pycbc_sample)

        os.makedirs(opts.output, exist_ok=True)

        if opts.condor_submit:
            if opts.seed is not None:
                raise NotImplementedError(
                    "--seed does not yet work with --condor-submit"
                )
            if opts.coinc_event_id:
                raise ValueError("must not set --coinc-event-id with --condor-submit")
            with subprocess.Popen(
                ["condor_submit"], text=True, stdin=subprocess.PIPE
            ) as proc:
                f = proc.stdin
                print(
                    """
                    accounting_group = ligo.dev.o4.cbc.pe.bayestar
                    on_exit_remove = (ExitBySignal == False) && (ExitCode == 0)
                    on_exit_hold = (ExitBySignal == True) || (ExitCode != 0)
                    on_exit_hold_reason = (ExitBySignal == True \
                        ? strcat("The job exited with signal ", ExitSignal) \
                        : strcat("The job exited with code ", ExitCode))
                    request_memory = 1000 MB
                    universe = vanilla
                    getenv = true
                    executable = /usr/bin/env
                    JobBatchName = BAYESTAR
                    environment = "OMP_NUM_THREADS=1"
                    """,
                    file=f,
                )
                print("error =", os.path.join(opts.output, "$(cid).err"), file=f)
                print("log =", os.path.join(opts.output, "$(cid).log"), file=f)
                print(
                    'arguments = "',
                    *(arg for arg in sys.argv if arg != "--condor-submit"),
                    '--coinc-event-id $(cid)"',
                    file=f,
                )
                print("queue cid in", *event_source, file=f)
            sys.exit(proc.returncode)

        if opts.coinc_event_id:
            event_source = {key: event_source[key] for key in opts.coinc_event_id}

        # Loop over all sngl_inspiral <-> sngl_inspiral coincs.
        for int_coinc_event_id, event in event_source.items():
            coinc_event_id = "coinc_event:coinc_event_id:{}".format(int_coinc_event_id)

            log.info("%s:preparing", coinc_event_id)

            epoch, sample_rate, epochs, snrs, responses, locations, horizons = (
                condition(
                    event,
                    waveform=opts.waveform,
                    f_low=opts.f_low,
                    enable_snr_series=opts.enable_snr_series,
                    f_high_truncate=opts.f_high_truncate,
                )
            )

            min_distance, max_distance, prior_distance_power, cosmology = (
                condition_prior(
                    horizons,
                    opts.min_distance,
                    opts.max_distance,
                    opts.prior_distance_power,
                    opts.cosmology,
                )
            )

            gmst = lal.GreenwichMeanSiderealTime(epoch)

            max_abs_t = 2 * snrs.data.shape[1] / sample_rate
            xmin = [0, -1, min_distance, -1, 0, 0]
            xmax = [2 * np.pi, 1, max_distance, 1, 2 * np.pi, 2 * max_abs_t]
            names = "ra dec distance inclination twopsi time".split()
            transformed_names = "ra sin_dec distance u twopsi time".split()
            forward_transforms = [
                identity,
                np.sin,
                identity,
                np.cos,
                identity,
                identity,
            ]
            reverse_transforms = [
                identity,
                np.arcsin,
                identity,
                np.arccos,
                identity,
                identity,
            ]
            kwargs = dict(
                min_distance=min_distance,
                max_distance=max_distance,
                prior_distance_power=prior_distance_power,
                cosmology=cosmology,
                gmst=gmst,
                sample_rate=sample_rate,
                epochs=epochs,
                snrs=snrs,
                responses=responses,
                locations=locations,
                horizons=horizons,
            )

            # Fix parameters
            for i, key in reversed(list(enumerate(["ra", "dec", "distance"]))):
                value = getattr(opts, key)
                if value is None:
                    continue

                if key in ["ra", "dec"]:
                    # FIXME: figure out a more elegant way to address different
                    # units in command line arguments and posterior samples
                    value = np.deg2rad(value)

                kwargs[transformed_names[i]] = forward_transforms[i](value)
                del (
                    xmin[i],
                    xmax[i],
                    names[i],
                    transformed_names[i],
                    forward_transforms[i],
                    reverse_transforms[i],
                )

            log.info("%s:sampling", coinc_event_id)

            # Run MCMC
            chain = ez_emcee(log_post, xmin, xmax, kwargs=kwargs, vectorize=True)

            # Transform back from sin_dec to dec and cos_inclination to
            # inclination
            for i, func in enumerate(reverse_transforms):
                chain[:, i] = func(chain[:, i])

            # Create Astropy table
            chain = Table(rows=chain, names=names, copy=False)

            log.info("%s:saving posterior samples", coinc_event_id)

            hdf5.write_samples(
                chain,
                os.path.join(opts.output, "{}.hdf5".format(int_coinc_event_id)),
                path="/bayestar/posterior_samples",
                overwrite=True,
            )



import numpy as np
from astropy.io.fits import Header
from astropy.table import Table
from astropy_healpix import HEALPix
from reproject import reproject_from_healpix

from ...moc import nest2uniq
from ..reproject_from_healpix_moc import reproject_from_healpix_moc


def test_reproject_from_healpix_moc():
    header = Header(
        {
            "NAXIS": 2,
            "NAXIS1": 180,
            "NAXIS2": 180,
            "CRPIX1": 90.5,
            "CRPIX2": 90.5,
            "CRVAL1": 180,
            "CRVAL2": 0,
            "CDELT1": 1,
            "CDELT2": 1,
            "CTYPE1": "RA---CAR",
            "CTYPE2": "DEC--CAR",
            "RADESYS": "ICRS",
        }
    )
    hpx = HEALPix(nside=32)
    map_flat = np.arange(hpx.npix)

    map_moc = Table(
        {
            "UNIQ": nest2uniq(np.int8(hpx.level), hpx.ring_to_nested(map_flat)),
            "IMG": map_flat,
        }
    )

    expected = reproject_from_healpix(
        (map_flat, "ICRS"), header, order="nearest-neighbor", nested=False
    )
    result = reproject_from_healpix_moc((map_moc, "ICRS"), header, "nearest-neighbor")

    np.testing.assert_array_equal(result, expected)



#
# Copyright (C) 2017-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Create a P-P plot to compare a posterior sample chain with a sky map."""

import re
from argparse import FileType

import numpy as np

from . import ArgumentParser
from .matplotlib import get_figure_parser


def fmt(x, sigfigs, force_scientific=False):
    """Round and format a number in scientific notation."""
    places = sigfigs - int(np.floor(np.log10(x)))
    x_rounded = np.around(x, places)
    if places <= 0 and not force_scientific:
        return "{:d}".format(int(x_rounded))
    else:
        s = ("{:." + str(sigfigs) + "e}").format(x_rounded)
        return re.sub(r"^(.*)e\+?(-?)0*(\d+)$", r"$\1 \\times 10^{\2\3}$", s)


def parser():
    parser = ArgumentParser(parents=[get_figure_parser()])
    parser.add_argument(
        "skymap",
        metavar="SKYMAP.fits[.gz]",
        type=FileType("rb"),
        help="FITS sky map file",
    )
    parser.add_argument(
        "samples",
        metavar="SAMPLES.hdf5",
        type=FileType("rb"),
        help="HDF5 posterior sample chain file",
    )
    parser.add_argument(
        "-m",
        "--max-points",
        type=int,
        help="Maximum number of posterior sample points [default: all of them]",
    )
    parser.add_argument(
        "-p",
        "--contour",
        default=[],
        nargs="+",
        type=float,
        metavar="PERCENT",
        help="Report the areas and volumes within the smallest contours "
        "containing this much probability.",
    )
    return parser


def main(args=None):
    with parser().parse_args(args) as args:
        # Late imports.
        from astropy.table import Table
        from matplotlib import pyplot as plt
        from scipy.interpolate import interp1d

        from .. import io
        from .. import plot as _  # noqa: F401
        from ..postprocess import find_injection_moc

        # Read input.
        skymap = io.read_sky_map(args.skymap.name, moc=True)
        chain = io.read_samples(args.samples.name)

        # If required, downselect to a smaller number of posterior samples.
        if args.max_points is not None:
            chain = Table(np.random.permutation(chain)[: args.max_points], copy=False)

        # Calculate P-P plot.
        contours = np.asarray(args.contour)
        result = find_injection_moc(
            skymap, chain["ra"], chain["dec"], chain["dist"], contours=1e-2 * contours
        )

        # Make Matplotlib figure.
        fig = plt.figure(figsize=(6, 6))
        ax = fig.add_subplot(111, projection="pp_plot")
        ax.add_diagonal()
        ax.add_series(result.searched_prob, label="R.A., Dec.")
        searched_area_func = interp1d(
            np.linspace(0, 1, len(chain)),
            np.sort(result.searched_area),
            bounds_error=False,
        )
        if "DISTMU" in skymap.colnames:
            ax.add_series(result.searched_prob_dist, label="Distance")
            ax.add_series(result.searched_prob_vol, label="Volume")
            searched_vol_func = interp1d(
                np.linspace(0, 1, len(chain)),
                np.sort(result.searched_vol),
                bounds_error=False,
            )
        for p, area, vol in zip(
            args.contour, result.contour_areas, result.contour_vols
        ):
            text = "{:g}%\n{} deg$^2$".format(p, fmt(area, 2))
            if "DISTMU" in skymap.colnames:
                text += "\n{} Mpc$^3$".format(fmt(vol, 2, force_scientific=True))
            ax.annotate(
                text,
                (1e-2 * p, 1e-2 * p),
                (0, -150),
                xycoords="data",
                textcoords="offset points",
                horizontalalignment="right",
                backgroundcolor="white",
                arrowprops=dict(
                    connectionstyle="bar,angle=0,fraction=0",
                    arrowstyle="-|>",
                    linewidth=2,
                    color="black",
                ),
            )
            area = searched_area_func(1e-2 * p)
            text = "{:g}%\n{} deg$^2$".format(p, fmt(area, 2))
            if "DISTMU" in skymap.colnames:
                vol = searched_vol_func(1e-2 * p)
                text += "\n{} Mpc$^3$".format(fmt(vol, 2, force_scientific=True))
            ax.annotate(
                text,
                (1e-2 * p, 1e-2 * p),
                (-75, 0),
                xycoords="data",
                textcoords="offset points",
                horizontalalignment="right",
                verticalalignment="center",
                backgroundcolor="white",
                arrowprops=dict(
                    connectionstyle="bar,angle=0,fraction=0",
                    arrowstyle="-|>",
                    linewidth=2,
                    color="black",
                ),
            )
        ax.set_xlabel("searched probability")
        ax.set_ylabel("cumulative fraction of posterior samples")
        ax.set_title(args.skymap.name)
        ax.legend()
        ax.grid()

        # Show or save output.
        args.output()



#
# Copyright (C) 2012-2025  Ethan Marx <emarx@mit.edu>
#                          Leo P. Singer <leo.singer@ligo.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from astropy import units as u
from astropy.coordinates import SkyCoord

try:
    from figaro.mixture import DPGMM
    from figaro.utils import get_priors
except ModuleNotFoundError as e:
    raise RuntimeError(
        "In order to use the DPGMM feature you must install `figaro`"
    ) from e

import numpy as np
from tqdm.auto import tqdm

from . import moc
from .coordinates import EigenFrame


class SkyDPGMM:
    def __init__(self, pts, **kwargs):
        self.frame = EigenFrame.for_coords(SkyCoord(*pts.T, unit=u.rad))

        # transform to eigenframe
        pts = self.transform(pts)

        # build DPGMM model
        bounds = [[0, 2 * np.pi], [-1, 1]]
        prior_pars = get_priors(bounds, pts)

        model = DPGMM(bounds, prior_pars=prior_pars)
        for s in tqdm(pts):
            model.add_new_point(s)

        self.model = model

    def transform(self, pts):
        pts = SkyCoord(*pts.T, unit=u.rad).transform_to(self.frame).spherical
        return np.column_stack((pts.lon.rad, np.sin(pts.lat.rad)))

    def __call__(self, pts):
        return self.model.pdf(self.transform(pts))

    def as_healpix(self, *args, **kwargs):
        return moc.bayestar_adaptive_grid(self, *args, **kwargs)



import platform
from itertools import chain, combinations, product

import astropy_healpix as ah
import matplotlib
from astropy.coordinates import ICRS, CartesianRepresentation, SkyCoord

matplotlib.use("agg")
import healpy as hp  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402
import numpy as np  # noqa: E402
import pytest  # noqa: E402
from astropy import units as u  # noqa: E402

from ..bayes_factor import plot_bayes_factor  # noqa: E402
from ..marker import reticle  # noqa: E402

skip_if_macos_arm64 = pytest.mark.skipif(
    platform.system() == "Darwin" and platform.machine() == "arm64",
    reason="Tick labels vary on macOS arm64",
)


def pp_plot():
    # Re-initialize the random seed to make the unit test repeatable
    np.random.seed(0)
    fig = plt.figure(figsize=(3, 3))
    ax = fig.add_subplot(111, projection="pp_plot")
    p_values = np.arange(1, 20) / 20
    return fig, ax, p_values


@pytest.fixture
def rcparams():
    with plt.rc_context({"figure.dpi": 72, "savefig.dpi": 72}):
        yield


@pytest.mark.mpl_image_compare(remove_text=True, tolerance=10)
def test_pp_plot_steps(rcparams):
    """Test P--P plot with drawstyle='steps'."""
    fig, ax, p_values = pp_plot()
    ax.add_confidence_band(len(p_values))
    ax.add_diagonal()
    ax.add_lightning(len(p_values), 20, drawstyle="steps")
    ax.add_series(p_values, drawstyle="steps")
    ax.add_worst(p_values)
    return fig


@pytest.mark.mpl_image_compare(remove_text=True, tolerance=10)
def test_pp_plot_lines(rcparams):
    """Test P--P plot with drawstyle='steps'."""
    fig, ax, p_values = pp_plot()
    ax.add_confidence_band(len(p_values))
    ax.add_diagonal()
    ax.add_lightning(len(p_values), 20, drawstyle="lines")
    ax.add_series(p_values, drawstyle="lines")
    ax.add_diagonal()
    ax.add_series(p_values)
    ax.add_worst(p_values)
    return fig


@pytest.mark.mpl_image_compare(remove_text=True, tolerance=10)
def test_pp_plot_default(rcparams):
    """Test P--P plot with drawstyle='steps'."""
    fig, ax, p_values = pp_plot()
    ax.add_confidence_band(len(p_values))
    ax.add_diagonal()
    ax.add_lightning(len(p_values), 20)
    ax.add_series(p_values)
    ax.add_worst(p_values)
    return fig


@pytest.mark.parametrize("proj", ["aitoff", "mollweide"])
@pytest.mark.parametrize("units", ["degrees", "hours"])
@pytest.mark.parametrize(
    "coordsys",
    [
        "astro",
        pytest.param("geo", marks=skip_if_macos_arm64),
        pytest.param("galactic", marks=skip_if_macos_arm64),
    ],
)
@pytest.mark.mpl_image_compare(remove_text=True, tolerance=1.5)
def test_allsky_axes(rcparams, coordsys, units, proj):
    """Test projection of a HEALPix image onto allsky axes, either
    in celestial or earth-fixed coordinates.
    """
    # Set up axes. (The obstime has an effect only for geographic axes.)
    fig = plt.figure(figsize=(6, 4))
    ax = fig.add_subplot(
        111,
        projection=coordsys + " " + units + " " + proj,
        obstime="2017-08-17T12:41:04.444458",
    )

    # Build a low-resolution example HEALPix sky map:
    # the value is equal to the right ascension.
    nside = 8
    npix = ah.nside_to_npix(nside)
    ra, dec = hp.pix2ang(nside, np.arange(npix), lonlat=True)
    img = np.sin(np.deg2rad(ra))

    # Plot, show grid, and return figure.
    ax.imshow_hpx((img, "ICRS"))
    ax.grid()
    return fig


@pytest.mark.mpl_image_compare(remove_text=True, tolerance=1.5)
def test_allsky_obstime(rcparams):
    fig = plt.figure(figsize=(6, 4))
    ax = fig.add_subplot(
        111, projection="geo degrees mollweide", obstime="2017-08-17T12:41:04.444458"
    )
    ax.grid()
    return fig


@pytest.mark.skipif(
    not (platform.system() == "Linux" and platform.machine == "x86_64"),
    reason="Tick label positions vary on different operating systems",
)
@pytest.mark.mpl_image_compare(remove_text=True, tolerance=1.5)
def test_globe_axes(rcparams):
    fig = plt.figure(figsize=(4, 4))
    ax = fig.add_axes(
        [0.2, 0.2, 0.6, 0.6], projection="astro globe", center="197.45d -23.38d"
    )
    ax.grid()
    return fig


@pytest.mark.mpl_image_compare(remove_text=True, tolerance=1.5)
def test_zoom_axes(rcparams):
    fig = plt.figure(figsize=(4, 4))
    ax = fig.add_axes(
        [0.2, 0.2, 0.6, 0.6],
        projection="astro zoom",
        center="197.45d -23.38d",
        radius="90 arcmin",
    )
    ax.scalebar((0.1, 0.1), 30 * u.arcmin)
    ax.grid()
    for key in ["ra", "dec"]:
        ax.coords[key].set_auto_axislabel(False)
    return fig


@pytest.mark.parametrize(
    "projection", ["astro zoom", "astro globe", "astro mollweide", "astro aitoff"]
)
def test_center_cartesian(projection):
    """Test that zoom axes accept coordinates in other representations."""
    fig = plt.figure()
    center = SkyCoord(0, 0, 1, representation_type=CartesianRepresentation)
    fig.add_axes([0.2, 0.2, 0.6, 0.6], projection=projection, center=center)


@pytest.mark.mpl_image_compare(remove_text=True, tolerance=1.5)
def test_reticle():
    which_list = [
        "".join(d)
        for d in chain.from_iterable(combinations("lrtb", n) for n in range(2, 5))
    ]
    inners = [0.0, 0.2, 0.4]
    outers = [0.8, 0.9, 1.0]
    angles = [0.0, 22.5, 45.0]
    args_list = list(product(inners, outers, angles))

    fig = plt.figure(figsize=(4, 4))
    ax = fig.add_axes([0, 0, 1, 1])
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    for args, x in zip(args_list, np.linspace(0.1, 0.9, len(args_list))):
        for which, y in zip(which_list, np.linspace(0.1, 0.9, len(which_list))):
            ax.plot(x, y, marker=reticle(*args, which=which))
    return fig


@pytest.mark.parametrize("proj", ["aitoff", "mollweide"])
@pytest.mark.parametrize("cent", ["197.45d -23.38d", None])
@pytest.mark.mpl_image_compare(remove_text=True, tolerance=1.5)
def test_center_projections(rcparams, proj, cent):
    fig = plt.figure(figsize=(4, 4))
    ax = fig.add_axes(111, projection=f"astro {proj}", center=cent)
    ax.grid()
    return fig


@pytest.mark.parametrize("frame", ["astro", "galactic"])
@pytest.mark.mpl_image_compare(remove_text=True, tolerance=1.5)
def test_center_galactic(rcparams, frame):
    """Regression test for https://github.com/lpsinger/ligo.skymap/issues/32"""
    center = SkyCoord(95.31043, -47.88693, frame=ICRS(), unit=u.deg)
    fig = plt.figure(figsize=(3, 3))
    ax = plt.axes(
        projection=f"{frame} degrees zoom",
        center=center,
        radius="30 deg",
        rotate="180 deg",
    )
    ax.scalebar((0.05, 0.05), 10 * u.deg).label()
    ax.compass(0.95, 0.05, 0.1)
    ax.grid()
    return fig


@pytest.mark.mpl_image_compare(tolerance=1.5)
def test_plot_bayes_factor():
    fig, ax = plot_bayes_factor(6.3, title="BAYESTAR is awesome")
    return fig



#
# Copyright (C) 2019-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Tools for progress bars"""

try:
    from billiard import Pool
except ImportError:
    from multiprocessing import Pool
from heapq import heappop, heappush
from operator import length_hint

import numpy as np
from tqdm.auto import tqdm

from .. import omp

__all__ = ("progress_map", "progress_map_vectorized")


class WrappedFunc:
    def __init__(self, func):
        self.func = func

    def __call__(self, i_args):
        i, args = i_args
        return i, self.func(*args)


class WrappedFuncWithLength:
    def __init__(self, func):
        self.func = func

    def __call__(self, i_args):
        i, args = i_args
        return len(args[0]), (i, self.func(*args))


def _get_total_estimate(*iterables):
    """Estimate total loop iterations for mapping over multiple iterables."""
    return min(length_hint(iterable) for iterable in iterables)


def _results_in_order(completed):
    """Put results back into order and yield them as quickly as they arrive."""
    heap = []
    current = 0
    for i_result in completed:
        i, result = i_result
        if i == current:
            yield result
            current += 1
            while heap and heap[0][0] == current:
                _, result = heappop(heap)
                yield result
                current += 1
        else:
            heappush(heap, i_result)
    assert not heap, "The heap must be empty"


_in_pool = False
"""Flag to prevent nested multiprocessing pools."""

_jobs = 1
_pool = None


def _init_process():
    """Disable OpenMP when using multiprocessing."""
    global _in_pool
    omp.num_threads = 1
    _in_pool = True


def progress_map(func, *iterables, jobs=1, **kwargs):
    """Map a function across iterables of arguments.

    Parameters
    ----------
    func : callable
        Function to evaluate.
    iterables
        Input argument iterables.
    jobs : int
        Number of subprocesses.
    kwargs : dict
        Additional keyword arguments passed to :obj:`tqdm.tqdm`.

    This is comparable to :meth:`astropy.utils.console.ProgressBar.map`, except
    that it is implemented using :mod:`tqdm` and so provides more detailed and
    accurate progress information.
    """
    global _jobs, _pool
    total = _get_total_estimate(*iterables)
    if _in_pool or jobs == 1:
        yield from tqdm(map(func, *iterables), total=total, **kwargs)
    else:
        if jobs != _jobs:
            if _pool is not None:
                _pool.close()
            _pool = Pool(jobs, _init_process)
            _jobs = jobs

        # Chunk size heuristic reproduced from
        # https://github.com/python/cpython/blob/v3.13.1/Lib/multiprocessing/pool.py#L481-L483.
        chunksize, extra = divmod(total, len(_pool._pool) * 4)
        if extra or chunksize == 0:
            chunksize += 1

        yield from _results_in_order(
            tqdm(
                _pool.imap_unordered(
                    WrappedFunc(func), enumerate(zip(*iterables)), chunksize=chunksize
                ),
                total=total,
                **kwargs,
            )
        )


def progress_map_vectorized(func, *iterables, jobs=1, nout=1, **kwargs):
    """Map a Numpy vectorized function across iterables of arguments.

    Optional parallelization is applied across the first axis of the
    iterables.

    Parameters
    ----------
    func : callable
        Function to evaluate.
    iterables
        Input argument arrays.
    jobs : int
        Number of subprocesses.
    nout : int
        Number of outputs of the function.
    kwargs : dict
        Additional keyword arguments passed to :obj:`tqdm.tqdm`.

    Notes
    -----
    All of the iterables must have equal length.
    """
    global _jobs, _pool
    if _in_pool or jobs == 1:
        return func(*iterables)
    else:
        # Check that all iterables have the same length
        total, *rest = (len(iterable) for iterable in iterables)
        for i, item in enumerate(rest):
            if item != total:
                raise ValueError(
                    f"Length of argument {i} ({item}) does not match length of argument 0 ({total})"
                )

        if jobs != _jobs:
            if _pool is not None:
                _pool.close()
            _pool = Pool(jobs, _init_process)
            _jobs = jobs

        # Chunk size heuristic reproduced from
        # https://github.com/python/cpython/blob/v3.13.1/Lib/multiprocessing/pool.py#L481-L483.
        chunks = min(total, len(_pool._pool) * 4)

        with tqdm(total=total, **kwargs) as progress:

            def update(n_result):
                n, result = n_result
                progress.update(n)
                return result

            result = list(
                _results_in_order(
                    update(n_result)
                    for n_result in _pool.imap_unordered(
                        WrappedFuncWithLength(func),
                        enumerate(
                            zip(
                                *(
                                    np.array_split(iterable, chunks)
                                    for iterable in iterables
                                )
                            )
                        ),
                        chunksize=1,
                    )
                )
            )

            if nout == 1:
                return np.concatenate(result)
            else:
                return np.column_stack(result)



#
# Copyright (C) 2014-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Construct a LIGO-LW XML power spectral density file for a network of
detectors by evaluating a model power noise sensitivity curve.
"""

import inspect
from argparse import SUPPRESS, FileType

from . import ArgumentParser, register_to_xmldoc, write_fileobj

psd_name_prefix = "SimNoisePSD"


def parser():
    import lal
    import lalsimulation

    # Get names of PSD functions.
    psd_names = sorted(
        name[len(psd_name_prefix) :]
        for name, func in inspect.getmembers(lalsimulation)
        if name.startswith(psd_name_prefix)
        and callable(func)
        and (
            "(double f) -> double" in func.__doc__
            or "(REAL8FrequencySeries psd, double flow) -> int" in func.__doc__
        )
    )

    parser = ArgumentParser()
    parser.add_argument(
        "-o",
        "--output",
        metavar="OUT.xml[.gz]",
        type=FileType("wb"),
        default="-",
        help="Name of output file [default: stdout]",
    )
    parser.add_argument(
        "--df",
        metavar="Hz",
        type=float,
        default=1.0,
        help="Frequency step size [default: %(default)s]",
    )
    parser.add_argument(
        "--f-max",
        metavar="Hz",
        type=float,
        default=2048.0,
        help="Maximum frequency [default: %(default)s]",
    )

    detector_group = parser.add_argument_group(
        "detector noise curves",
        "Options to select noise curves for detectors.\n\n"
        "All detectors support the following options:\n\n" + "\n".join(psd_names),
    )

    scale_group = parser.add_argument_group(
        "detector scaling",
        "Options to apply scale factors to noise curves for detectors.\n"
        "For example, a scale factor of 2 means that the amplitude spectral\n"
        "density is multiplied by 1/2 so that the range is multiplied by a 2.",
    )

    # Add options for individual detectors
    for detector in lal.CachedDetectors:
        name = detector.frDetector.name
        prefix = detector.frDetector.prefix
        detector_group.add_argument(
            "--" + prefix,
            choices=psd_names,
            metavar="func",
            default=SUPPRESS,
            help="PSD function for {0} detector".format(name),
        )
        scale_group.add_argument(
            "--" + prefix + "-scale",
            type=float,
            default=SUPPRESS,
            help="Scale range for {0} detector".format(name),
        )

    return parser


def main(args=None):
    p = parser()
    with p.parse_args(args) as opts:
        import lal.series
        import lalsimulation
        import numpy as np

        from ..bayestar.filter import vectorize_swig_psd_func

        # Add basic options.

        psds = {}

        n = int(opts.f_max // opts.df)
        f = np.arange(n) * opts.df

        detectors = [d.frDetector.prefix for d in lal.CachedDetectors]

        for detector in detectors:
            psd_name = getattr(opts, detector, None)
            if psd_name is None:
                continue
            scale = 1 / np.square(getattr(opts, detector + "_scale", 1.0))
            func = getattr(lalsimulation, psd_name_prefix + psd_name)
            series = lal.CreateREAL8FrequencySeries(
                psd_name, 0, 0, opts.df, lal.SecondUnit, n
            )
            if "(double f) -> double" in func.__doc__:
                series.data.data = vectorize_swig_psd_func(psd_name_prefix + psd_name)(
                    f
                )
            else:
                func(series, 0.0)

                # Find indices of first and last nonzero samples.
                nonzero = np.flatnonzero(series.data.data)
                # FIXME: int cast seems to be needed on old versions of Numpy
                first_nonzero = int(nonzero[0])
                last_nonzero = int(nonzero[-1])

                # Truncate
                series = lal.CutREAL8FrequencySeries(
                    series, first_nonzero, last_nonzero - first_nonzero + 1
                )
                series.f0 = first_nonzero * series.deltaF

                series.name = psd_name
            series.data.data *= scale
            psds[detector] = series

        xmldoc = lal.series.make_psd_xmldoc(psds, encoding="base64")
        register_to_xmldoc(xmldoc, p, opts)
        write_fileobj(xmldoc, opts.output)



# FIXME: Remove this file if https://github.com/willvousden/ptemcee/pull/6
# is merged
import numpy as np
import ptemcee.sampler

__all__ = ("Sampler",)


class VectorLikePriorEvaluator(ptemcee.sampler.LikePriorEvaluator):
    def __call__(self, x):
        s = x.shape
        x = x.reshape((-1, x.shape[-1]))

        lp = self.logp(x, *self.logpargs, **self.logpkwargs)
        if np.any(np.isnan(lp)):
            raise ValueError("Prior function returned NaN.")

        ll = np.empty_like(lp)
        bad = lp == -np.inf
        ll[bad] = 0
        ll[~bad] = self.logl(x[~bad], *self.loglargs, **self.loglkwargs)
        if np.any(np.isnan(ll)):
            raise ValueError("Log likelihood function returned NaN.")

        return ll.reshape(s[:-1]), lp.reshape(s[:-1])


class Sampler(ptemcee.sampler.Sampler):
    """Patched version of :class:`ptemcee.Sampler` that supports the
    `vectorize` option of :class:`emcee.EnsembleSampler`.
    """

    def __init__(
        self,
        nwalkers,
        dim,
        logl,
        logp,  # noqa: N803
        ntemps=None,
        Tmax=None,
        betas=None,  # noqa: N803
        threads=1,
        pool=None,
        a=2.0,
        loglargs=[],
        logpargs=[],
        loglkwargs={},
        logpkwargs={},
        adaptation_lag=10000,
        adaptation_time=100,
        random=None,
        vectorize=False,
    ):
        super().__init__(
            nwalkers,
            dim,
            logl,
            logp,
            ntemps=ntemps,
            Tmax=Tmax,
            betas=betas,
            threads=threads,
            pool=pool,
            a=a,
            loglargs=loglargs,
            logpargs=logpargs,
            loglkwargs=loglkwargs,
            logpkwargs=logpkwargs,
            adaptation_lag=adaptation_lag,
            adaptation_time=adaptation_time,
            random=random,
        )
        self._vectorize = vectorize
        if vectorize:
            self._likeprior = VectorLikePriorEvaluator(
                logl, logp, loglargs, logpargs, loglkwargs, logpkwargs
            )

    def _evaluate(self, ps):
        if self._vectorize:
            return self._likeprior(ps)
        else:
            return super().evaluate(ps)



#
# Copyright (C) 2013-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Produce GW sky maps for all coincidences in search pipeline output database
in LIGO-LW XML, LIGO-LW SQLite, or PyCBC HDF5 format.

The distance prior is controlled by the ``--prior-distance-power`` argument.
If you set ``--prior-distance-power`` to k, then the distance prior is
proportional to r^k. The default is 2, uniform in volume.

If the ``--min-distance`` argument is omitted, it defaults to zero. If the
``--max-distance argument`` is omitted, it defaults to the SNR=4 horizon
distance of the most sensitive detector.

A FITS file is created for each sky map, having a filename of the form
``X.fits`` where X is the integer LIGO-LW row ID of the coinc. The ``OBJECT``
card in the FITS header is also set to the integer row ID.
"""

from argparse import FileType

from . import (
    ArgumentParser,
    get_mcmc_parser,
    get_posterior_parser,
    get_random_parser,
    get_waveform_parser,
)

ROW_ID_COMMENT = [
    "",
    "The integer value in the OBJECT card in this FITS header is a row ID",
    "that refers to a coinc_event table row in the input LIGO-LW document.",
    "",
]


def parser():
    parser = ArgumentParser(
        parents=[
            get_waveform_parser(),
            get_posterior_parser(),
            get_mcmc_parser(),
            get_random_parser(),
        ]
    )
    parser.add_argument(
        "-d",
        "--disable-detector",
        metavar="X1",
        type=str,
        nargs="+",
        help="disable certain detectors",
    )
    parser.add_argument(
        "--keep-going",
        "-k",
        default=False,
        action="store_true",
        help="Keep processing events if a sky map fails to converge",
    )
    parser.add_argument(
        "input",
        metavar="INPUT.{hdf,xml,xml.gz,sqlite}",
        default="-",
        nargs="+",
        type=FileType("rb"),
        help="Input LIGO-LW XML file, SQLite file, or PyCBC HDF5 files. "
        "For PyCBC, you must supply the coincidence file "
        '(e.g. "H1L1-HDFINJFIND.hdf" or "H1L1-STATMAP.hdf"), '
        "the template bank file (e.g. H1L1-BANK2HDF.hdf), "
        "the single-detector merged PSD files "
        '(e.g. "H1-MERGE_PSDS.hdf" and "L1-MERGE_PSDS.hdf"), '
        "and the single-detector merged trigger files "
        '(e.g. "H1-HDF_TRIGGER_MERGE.hdf" and '
        '"L1-HDF_TRIGGER_MERGE.hdf"), '
        "in any order.",
    )
    parser.add_argument(
        "--pycbc-sample", default="foreground", help="(PyCBC only) sample population"
    )
    parser.add_argument(
        "--coinc-event-id",
        type=int,
        nargs="*",
        help="run on only these specified events",
    )
    parser.add_argument("--output", "-o", default=".", help="output directory")
    parser.add_argument(
        "--condor-submit",
        action="store_true",
        help="submit to Condor instead of running locally",
    )
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        import logging

        log = logging.getLogger("BAYESTAR")

        # BAYESTAR imports.
        # Other imports.
        import os
        import subprocess
        import sys

        from .. import omp
        from ..bayestar import localize
        from ..io import events, fits

        log.info("Using %d OpenMP thread(s)", omp.num_threads)

        # Read coinc file.
        log.info("%s:reading input files", ",".join(file.name for file in opts.input))
        event_source = events.open(*opts.input, sample=opts.pycbc_sample)

        if opts.disable_detector:
            event_source = events.detector_disabled.open(
                event_source, opts.disable_detector
            )

        os.makedirs(opts.output, exist_ok=True)

        if opts.condor_submit:
            if opts.seed is not None:
                raise NotImplementedError(
                    "--seed does not yet work with --condor-submit"
                )
            if opts.coinc_event_id:
                raise ValueError("must not set --coinc-event-id with --condor-submit")
            with subprocess.Popen(
                ["condor_submit"], text=True, stdin=subprocess.PIPE
            ) as proc:
                f = proc.stdin
                print(
                    """
                    accounting_group = ligo.dev.o4.cbc.pe.bayestar
                    on_exit_remove = (ExitBySignal == False) && (ExitCode == 0)
                    on_exit_hold = (ExitBySignal == True) || (ExitCode != 0)
                    on_exit_hold_reason = (ExitBySignal == True \
                        ? strcat("The job exited with signal ", ExitSignal) \
                        : strcat("The job exited with code ", ExitCode))
                    request_memory = 2000 MB
                    request_disk = 100 MB
                    universe = vanilla
                    getenv = true
                    executable = /usr/bin/env
                    JobBatchName = BAYESTAR
                    environment = "OMP_NUM_THREADS=1"
                    """,
                    file=f,
                )
                print("error =", os.path.join(opts.output, "$(cid).err"), file=f)
                print(
                    'arguments = "',
                    *(arg for arg in sys.argv if arg != "--condor-submit"),
                    '--coinc-event-id $(cid)"',
                    file=f,
                )
                print("queue cid in", *event_source, file=f)
            sys.exit(proc.returncode)

        if opts.coinc_event_id:
            event_source = {key: event_source[key] for key in opts.coinc_event_id}

        count_sky_maps_failed = 0

        # Loop over all sngl_inspiral <-> sngl_inspiral coincs.
        for coinc_event_id, event in event_source.items():
            # Loop over sky localization methods
            log.info("%d:computing sky map", coinc_event_id)
            if opts.chain_dump:
                chain_dump = f"{coinc_event_id}.hdf5"
            else:
                chain_dump = None
            try:
                sky_map = localize(
                    event,
                    opts.waveform,
                    opts.f_low,
                    opts.min_distance,
                    opts.max_distance,
                    opts.prior_distance_power,
                    opts.cosmology,
                    mcmc=opts.mcmc,
                    chain_dump=chain_dump,
                    enable_snr_series=opts.enable_snr_series,
                    f_high_truncate=opts.f_high_truncate,
                    rescale_loglikelihood=opts.rescale_loglikelihood,
                )
                sky_map.meta["objid"] = coinc_event_id
                sky_map.meta["comment"] = ROW_ID_COMMENT

            except (ArithmeticError, ValueError):
                log.exception("%d:sky localization failed", coinc_event_id)
                count_sky_maps_failed += 1
                if not opts.keep_going:
                    raise
            else:
                log.info("%d:saving sky map", coinc_event_id)
                filename = f"{coinc_event_id}.fits"
                fits.write_sky_map(
                    os.path.join(opts.output, filename), sky_map, nest=True
                )

        if count_sky_maps_failed > 0:
            raise RuntimeError(
                "{0} sky map{1} did not converge".format(
                    count_sky_maps_failed, "s" if count_sky_maps_failed > 1 else ""
                )
            )



#
# Copyright (C) 2016-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Specialized markers."""

import numpy as np
from matplotlib.markers import MarkerStyle
from matplotlib.path import Path
from matplotlib.transforms import Affine2D

__all__ = ("earth", "sun", "moon", "reticle")


earth = Path.unit_circle()
earth = MarkerStyle(
    Path(
        np.concatenate((earth.vertices, [[-1, 0], [1, 0], [0, -1], [0, 1]])),
        np.concatenate((earth.codes, [Path.MOVETO, Path.LINETO] * 2)),
    ),
    fillstyle="none",
)
earth.__doc__ = """
The Earth symbol (circle and cross).

Examples
--------
.. plot::
   :context: reset
   :include-source:
   :align: center

    from matplotlib import pyplot as plt
    from ligo.skymap.plot.marker import earth

    plt.plot(0, 0, marker=earth, markersize=20, markeredgewidth=2)

"""


sun = Path.unit_circle()
sun = MarkerStyle(
    Path(
        np.concatenate((sun.vertices, [[0, 0], [1e-3, 0]])),
        np.concatenate((sun.codes, [Path.MOVETO, Path.LINETO])),
    ),
    capstyle="round",
    joinstyle="round",
    fillstyle="none",
)
sun.__doc__ = """
The Sun symbol (circle and dot).

Examples
--------
.. plot::
   :context: reset
   :include-source:
   :align: center

    from matplotlib import pyplot as plt
    from ligo.skymap.plot.marker import sun

    plt.plot(0, 0, marker=sun, markersize=20, markeredgewidth=2)

"""


def moon(phase, shadow=False):
    """Create a marker in the shape of the Moon.

    Parameters
    ----------
    phase : float
        Lunar phase in degrees between -180 and 180.
    shadow : bool
        If set, then the shadowed portion of the Moon is included in the
        marker, and its fill color can be set independently using the
        ``markerfacecoloralt`` keyword argument for
        :meth:`~matplotlib.axes.Axes.plot` (see
        :ref:`matplotlib:marker_fill_styles`).

    Returns
    -------
    markerstyle : matplotlib.markers.MarkerStyle

    Examples
    --------

    .. plot::
       :context: reset
       :include-source:
       :align: center

        from matplotlib import pyplot as plt
        from matplotlib.ticker import MultipleLocator
        import numpy as np
        from ligo.skymap.plot.marker import moon

        d_phase = 30
        phases = np.arange(-180, 180 + d_phase, d_phase)

        fig, ax = plt.subplots(figsize=(8, 3), tight_layout=True)
        ax.xaxis.set_major_locator(MultipleLocator(d_phase))
        for phase in phases:
            ax.plot(phase, 4, ms=20, marker=moon(phase, shadow=False), mfc="none", mec="black")
            ax.plot(phase, 3, ms=20, marker=moon(phase, shadow=False), mfc="goldenrod", mec="none")
            ax.plot(phase, 2, ms=20, marker=moon(phase, shadow=False), mfc="goldenrod", mec="k")
            ax.plot(phase, 1, ms=20, marker=moon(phase, shadow=True), mfc="goldenrod", mfcalt="gray", mec="none")
            ax.plot(phase, 0, ms=20, marker=moon(phase, shadow=True), mfc="goldenrod", mfcalt="gray", mec="black")
        ax.set_yticks(
            [0, 1, 2, 3, 4],
            ["shadow, fill, stroke", "shadow, fill", "fill, stroke", "fill", "stroke"],
        )
        ax.set_ylim(-0.5, 4.5)

    """  # noqa: E501
    angle = np.deg2rad(90 - phase)
    sign = np.sign(np.cos(angle))
    arc = Path.arc(90, 270, 9)

    path1 = arc.transformed(Affine2D().scale(sign * np.sin(angle), 1))
    path2 = arc.transformed(Affine2D().scale(-sign, 1))
    path3 = arc.transformed(Affine2D().scale(sign, 1))

    light_path = Path(
        np.concatenate((path1.vertices, path2.vertices[::-1])),
        np.concatenate((path1.codes, path2.codes[:0:-1], [Path.CLOSEPOLY])),
    )
    dark_path = Path(
        np.concatenate((path1.vertices, path3.vertices[::-1])),
        np.concatenate((path1.codes, path3.codes[:0:-1], [Path.CLOSEPOLY])),
    )

    markerstyle = MarkerStyle(light_path, joinstyle="miter")
    if shadow:
        markerstyle._alt_path = dark_path
        markerstyle._alt_transform = markerstyle._transform
    return markerstyle


def reticle(inner=0.5, outer=1.0, angle=0.0, which="lrtb"):
    """Create a reticle or crosshairs marker.

    Parameters
    ----------
    inner : float
        Distance from the origin to the inside of the crosshairs.
    outer : float
        Distance from the origin to the outside of the crosshairs.
    angle : float
        Rotation in degrees; 0 for a '+' orientation and 45 for 'x'.

    Returns
    -------
    path : `matplotlib.path.Path`
        The new marker path, suitable for passing to Matplotlib functions
        (e.g., `plt.plot(..., marker=reticle())`)

    Examples
    --------
    .. plot::
       :context: reset
       :include-source:
       :align: center

        from matplotlib import pyplot as plt
        from ligo.skymap.plot.marker import reticle

        markers = [reticle(inner=0),
                   reticle(which='lt'),
                   reticle(which='lt', angle=45)]

        fig, ax = plt.subplots(figsize=(6, 2))
        ax.set_xlim(-0.5, 2.5)
        ax.set_ylim(-0.5, 0.5)
        for x, marker in enumerate(markers):
            ax.plot(x, 0, markersize=20, markeredgewidth=2, marker=marker)

    """
    angle = np.deg2rad(angle)
    x = np.cos(angle)
    y = np.sin(angle)
    rotation = [[x, y], [-y, x]]
    vertdict = {"l": [-1, 0], "r": [1, 0], "b": [0, -1], "t": [0, 1]}
    verts = [vertdict[direction] for direction in which]
    codes = [Path.MOVETO, Path.LINETO] * len(verts)
    verts = np.dot(verts, rotation)
    verts = np.swapaxes([inner * verts, outer * verts], 0, 1).reshape(-1, 2)
    return Path(verts, codes)



#
# Copyright (C) 2012-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Angle utilities."""

import numpy as np

__all__ = (
    "reference_angle",
    "reference_angle_deg",
    "wrapped_angle",
    "wrapped_angle_deg",
)


def reference_angle(a):
    """Convert an angle to a reference angle between -pi and pi.

    Examples
    --------

    >>> reference_angle(1.5 * np.pi)
    array(-1.57079633)
    """
    a = np.mod(a, 2 * np.pi)
    return np.where(a <= np.pi, a, a - 2 * np.pi)


def reference_angle_deg(a):
    """Convert an angle to a reference angle between -180 and 180 degrees.

    Examples
    --------

    >>> reference_angle_deg(270.)
    array(-90.)
    """
    a = np.mod(a, 360)
    return np.where(a <= 180, a, a - 360)


def wrapped_angle(a):
    """Convert an angle to a reference angle between 0 and 2*pi.

    Examples
    --------

    >>> print(wrapped_angle(3 * np.pi))
    3.141592653589793
    """
    return np.mod(a, 2 * np.pi)


def wrapped_angle_deg(a):
    """Convert an angle to a reference angle between 0 and 2*pi.

    Examples
    --------

    >>> print(wrapped_angle_deg(540.))
    180.0
    """
    return np.mod(a, 360)



#
# Copyright (C) 2013-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""
Calculate summary statistics for a batch of sky maps.

Under the hood, this script is little more than a command-line interface for
the :mod:`ligo.skymap.postprocess.crossmatch` module.

The filenames of the sky maps may be provided as positional command line
arguments, and may also be provided as globs (such as ``*.fits.gz``). If
supplied with the optional ``--database`` argument, then also match sky maps
with injections from an inspinjfind-style sqlite database.

All angular separations are in degrees, all areas are in square degrees, and
all volumes are in cubic megaparsecs. The output is written as tab-separated
values with the following columns:

+-----------------------------------------------------------------------------+
| **From the search pipeline database**                                       |
+------------------------+----------------------------------------------------+
| ``coinc_event_id``     | event identifier                                   |
+------------------------+----------------------------------------------------+
| ``simulation_id``      | injection identifier                               |
+------------------------+----------------------------------------------------+
| ``far``                | false alarm rate                                   |
+------------------------+----------------------------------------------------+
| ``snr``                | signal to noise ratio                              |
+------------------------+----------------------------------------------------+
| **Injection finding**                                                       |
+------------------------+----------------------------------------------------+
| ``searched_area``      | area of 2D credible region containing the true sky |
|                        | location                                           |
+------------------------+----------------------------------------------------+
| ``searched_prob``      | probability in that 2D credible region             |
+------------------------+----------------------------------------------------+
| ``searched_prob_dist`` | marginal distance CDF at the true distance         |
+------------------------+----------------------------------------------------+
| ``searched_vol``       | volume of 3D credible region containing the true   |
|                        | position                                           |
+------------------------+----------------------------------------------------+
| ``searched_prob_vol``  | probability contained in that volume               |
+------------------------+----------------------------------------------------+
| ``offset``             | angular separation between the maximum             |
|                        | *a posteriori* position and the true sky position  |
+------------------------+----------------------------------------------------+
| **Additional metadata from the sky maps**                                   |
+------------------------+----------------------------------------------------+
| ``runtime``            | wall clock run time to generate sky map            |
+------------------------+----------------------------------------------------+
| ``distmean``           | mean *a posteriori* distance                       |
+------------------------+----------------------------------------------------+
| ``diststd``            | *a posteriori* standard deviation of distance      |
+------------------------+----------------------------------------------------+
| ``log_bci``            | natural log Bayes factor, coherent vs. incoherent  |
+------------------------+----------------------------------------------------+
| ``log_bsn``            | natural log Bayes factor, signal vs. noise         |
+------------------------+----------------------------------------------------+
| **Credible levels** (if ``--area`` or ``--contour`` options present)        |
+------------------------+----------------------------------------------------+
| ``area(P)``            | area of the *P* percent 2D credible region         |
+------------------------+----------------------------------------------------+
| ``prob(A)``            | probability contained within the 2D credible level |
|                        | of area *A*                                        |
+------------------------+----------------------------------------------------+
| ``dist(P)``            | distance for a cumulative marginal probability of  |
|                        | *P* percent                                        |
+------------------------+----------------------------------------------------+
| ``vol(P)``             | volume of the *P* percent 3D credible region       |
+------------------------+----------------------------------------------------+
| **Modes** (if ``--modes`` option is present)                                |
+------------------------+----------------------------------------------------+
| ``searched_modes``     | number of simply connected figures in the 2D       |
|                        | credible region containing the true sky location   |
+------------------------+----------------------------------------------------+
| ``modes(P)``           | number of simply connected figures in the *P*      |
|                        | percent 2D credible region                         |
+------------------------+----------------------------------------------------+

"""

import sys
from argparse import FileType
from functools import partial

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord

from ..io import fits
from ..postprocess import crossmatch
from . import ArgumentParser, SQLiteType


def parser():
    parser = ArgumentParser()
    parser.add_argument(
        "-o",
        "--output",
        metavar="OUT.dat",
        type=FileType("w"),
        default="-",
        help="Name of output file",
    )
    parser.add_argument(
        "-j",
        "--jobs",
        type=int,
        default=1,
        const=None,
        nargs="?",
        help="Number of threads",
    )
    parser.add_argument(
        "-p",
        "--contour",
        default=[],
        nargs="+",
        type=float,
        metavar="PERCENT",
        help="Report the area of the smallest contour and the number of modes "
        "containing this much probability.",
    )
    parser.add_argument(
        "-a",
        "--area",
        default=[],
        nargs="+",
        type=float,
        metavar="DEG2",
        help="Report the largest probability contained within any region "
        "of this area in square degrees. Can be repeated multiple times.",
    )
    parser.add_argument(
        "--modes", action="store_true", help="Compute number of disjoint modes"
    )
    parser.add_argument(
        "-d",
        "--database",
        type=SQLiteType("r"),
        metavar="DB.sqlite",
        help="Input SQLite database from search pipeline",
    )
    parser.add_argument(
        "fitsfilenames",
        metavar="GLOB.fits[.gz]",
        nargs="+",
        action="glob",
        help="Input FITS filenames and/or globs",
    )
    parser.add_argument(
        "--cosmology",
        action="store_true",
        help="Report volume localizations as comoving volumes.",
    )
    return parser


def process(fitsfilename, db, contours, modes, areas, cosmology):
    sky_map = fits.read_sky_map(fitsfilename, moc=True)

    coinc_event_id = sky_map.meta.get("objid")
    try:
        runtime = sky_map.meta["runtime"]
    except KeyError:
        runtime = float("nan")

    contour_pvalues = 0.01 * np.asarray(contours)

    if db is None:
        simulation_id = true_ra = true_dec = true_dist = far = snr = None
    else:
        row = db.execute(
            """
            SELECT DISTINCT sim.simulation_id AS simulation_id,
            sim.longitude AS ra, sim.latitude AS dec, sim.distance AS distance,
            ci.combined_far AS far, ci.snr AS snr
            FROM coinc_event_map AS cem1 INNER JOIN coinc_event_map AS cem2
            ON (cem1.coinc_event_id = cem2.coinc_event_id)
            INNER JOIN sim_inspiral AS sim
            ON (cem1.event_id = sim.simulation_id)
            INNER JOIN coinc_inspiral AS ci
            ON (cem2.event_id = ci.coinc_event_id)
            WHERE cem1.table_name = 'sim_inspiral'
            AND cem2.table_name = 'coinc_event' AND cem2.event_id = ?
            """,
            (coinc_event_id,),
        ).fetchone()
        if row is None:
            return None
        simulation_id, true_ra, true_dec, true_dist, far, snr = row

    if true_ra is None or true_dec is None:
        true_coord = None
    elif true_dist is None:
        true_coord = SkyCoord(true_ra * u.rad, true_dec * u.rad)
    else:
        true_coord = SkyCoord(true_ra * u.rad, true_dec * u.rad, true_dist * u.Mpc)

    (
        searched_area,
        searched_prob,
        offset,
        searched_modes,
        contour_areas,
        area_probs,
        contour_modes,
        searched_prob_dist,
        contour_dists,
        searched_vol,
        searched_prob_vol,
        contour_vols,
        probdensity,
        probdensity_vol,
    ) = crossmatch(
        sky_map,
        true_coord,
        contours=contour_pvalues,
        areas=areas,
        modes=modes,
        cosmology=cosmology,
    )

    if snr is None:
        snr = np.nan
    if far is None:
        far = np.nan
    distmean = sky_map.meta.get("distmean", np.nan)
    diststd = sky_map.meta.get("diststd", np.nan)
    log_bci = sky_map.meta.get("log_bci", np.nan)
    log_bsn = sky_map.meta.get("log_bsn", np.nan)

    ret = [coinc_event_id]
    if db is not None:
        ret += [
            simulation_id,
            far,
            snr,
            searched_area,
            searched_prob,
            searched_prob_dist,
            searched_vol,
            searched_prob_vol,
            offset,
        ]
    ret += [runtime, distmean, diststd, log_bci, log_bsn]
    ret += contour_areas + area_probs + contour_dists + contour_vols
    if modes:
        if db is not None:
            ret += [searched_modes]
        ret += contour_modes
    return ret


def main(args=None):
    p = parser()
    with p.parse_args(args) as opts:
        from ..util.progress import progress_map

        if args is None:
            print("#", *sys.argv, file=opts.output)
        else:
            print("#", p.prog, *args, file=opts.output)

        colnames = ["coinc_event_id"]
        if opts.database is not None:
            colnames += [
                "simulation_id",
                "far",
                "snr",
                "searched_area",
                "searched_prob",
                "searched_prob_dist",
                "searched_vol",
                "searched_prob_vol",
                "offset",
            ]
        colnames += ["runtime", "distmean", "diststd", "log_bci", "log_bsn"]
        colnames += ["area({0:g})".format(_) for _ in opts.contour]
        colnames += ["prob({0:g})".format(_) for _ in opts.area]
        colnames += ["dist({0:g})".format(_) for _ in opts.contour]
        colnames += ["vol({0:g})".format(_) for _ in opts.contour]
        if opts.modes:
            if opts.database is not None:
                colnames += ["searched_modes"]
            colnames += ["modes({0:g})".format(p) for p in opts.contour]
        print(*colnames, sep="\t", file=opts.output)

        func = partial(
            process,
            db=opts.database,
            contours=opts.contour,
            modes=opts.modes,
            areas=opts.area,
            cosmology=opts.cosmology,
        )
        for record in progress_map(func, opts.fitsfilenames, jobs=opts.jobs):
            if record is not None:
                print(*record, sep="\t", file=opts.output)



"""Pytest plugin for adding the number of OpenMP threads to the test report."""


def pytest_addoption(parser):
    doc = "Show the value of omp_get_num_threads() in the test header"
    group = parser.getgroup("OpenMP options")
    group.addoption("--omp-get-num-threads", action="store_true", help=doc)
    parser.addini("omp_get_num_threads", type="bool", help=doc)


def pytest_report_header(config):
    key = "omp_get_num_threads"
    if config.getoption(key) or config.getini(key):
        from ... import omp

        return f"omp_get_num_threads(): {omp.num_threads}\n"



from os import getpid
from time import sleep

import numpy as np
import pytest

from ..progress import progress_map, progress_map_vectorized


def func(x):
    sleep(np.random.uniform(0, 0.1))
    return np.square(x)


def func2(x):
    sleep(np.random.uniform(0, 0.1))
    return np.cos(x), np.sin(x)


@pytest.mark.parametrize("jobs", [1, 8, None])
def test_map(jobs):
    x = np.arange(20)
    result = list(progress_map(func, x, jobs=jobs))
    np.testing.assert_array_equal(result, np.square(x))


def map0(_):
    return getpid()


def map1(_):
    return list(progress_map(map0, range(8), jobs=8))


def map2():
    return sum(progress_map(map1, range(8), jobs=8), [])


def test_no_nested_pools():
    """Test that parallelism is disabled in nested calls to progress_map."""
    assert len(set(map2())) <= 8


@pytest.mark.parametrize("jobs", [1, 8, None])
def test_indefinite(jobs):
    """Test iteration over a collection of indefinite length."""
    assert list(progress_map(np.square, (i for i in range(3)), jobs=jobs)) == [0, 1, 4]


@pytest.mark.parametrize("jobs", [1, 8, None])
def test_map_vectorized(jobs):
    x = np.arange(20)
    result = progress_map_vectorized(func, x, jobs=jobs)
    np.testing.assert_array_equal(result, np.square(x))


@pytest.mark.parametrize("jobs", [1, 8, None])
def test_map_vectorized_nout(jobs):
    x = np.arange(20)
    result1, result2 = progress_map_vectorized(func2, x, jobs=jobs, nout=2)
    np.testing.assert_array_equal(result1, np.cos(x))
    np.testing.assert_array_equal(result2, np.sin(x))



#
# Copyright (C) 2015-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Create a contours for the credible levels of an all-sky probability map.
The input is a HEALPix probability map.
The output is a GeoJSON FeatureCollection (http://geojson.org/).
"""

from argparse import FileType

from . import ArgumentParser


def parser():
    parser = ArgumentParser()
    parser.add_argument(
        "-o",
        "--output",
        metavar="FILE.geojson",
        default="-",
        type=FileType("w"),
        help="output file [default: stdout]",
    )
    parser.add_argument(
        "--contour",
        metavar="PERCENT",
        type=float,
        nargs="+",
        required=True,
        help="plot contour enclosing this percentage of probability mass",
    )
    parser.add_argument(
        "-i",
        "--interpolate",
        choices="nearest nested bilinear".split(),
        default="nearest",
        help="resampling interpolation method",
    )
    parser.add_argument(
        "-s", "--simplify", action="store_true", help="simplify contour paths"
    )
    parser.add_argument(
        "-n",
        "--nside",
        metavar="NSIDE",
        type=int,
        help="optionally resample to the specified resolution "
        " before generating contours",
    )
    parser.add_argument(
        "input",
        metavar="INPUT.fits[.gz]",
        type=FileType("rb"),
        default="-",
        nargs="?",
        help="Input FITS file",
    )
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        import json

        import healpy as hp
        import numpy as np

        from .. import postprocess
        from ..io import fits

        # Read input file
        prob, _ = fits.read_sky_map(opts.input.name, nest=True)

        # Resample if requested
        if opts.nside is not None and opts.interpolate in ("nearest", "nested"):
            prob = hp.ud_grade(prob, opts.nside, order_in="NESTED", power=-2)
        elif opts.nside is not None and opts.interpolate == "bilinear":
            prob = postprocess.smooth_ud_grade(prob, opts.nside, nest=True)
        if opts.interpolate == "nested":
            prob = postprocess.interpolate_nested(prob, nest=True)

        # Find credible levels
        i = np.flipud(np.argsort(prob))
        cumsum = np.cumsum(prob[i])
        cls = np.empty_like(prob)
        cls[i] = cumsum * 100

        # Generate contours
        paths = list(
            postprocess.contour(
                cls, opts.contour, nest=True, degrees=True, simplify=opts.simplify
            )
        )

        json.dump(
            {
                "type": "FeatureCollection",
                "features": [
                    {
                        "type": "Feature",
                        "properties": {"credible_level": contour},
                        "geometry": {"type": "MultiLineString", "coordinates": path},
                    }
                    for contour, path in zip(opts.contour, paths)
                ],
            },
            opts.output,
        )



import numpy as np
import pytest

from ... import distance, io
from . import run_entry_point


@pytest.fixture
def skymap(tmpdir):
    npix = 12
    filename = str(tmpdir / "skymap.fits")
    distmean = 100.0
    diststd = 20.0
    distmu, distsigma, distnorm = distance.moments_to_parameters(distmean, diststd)
    io.write_sky_map(
        filename,
        (
            np.tile(1 / npix, npix),
            np.tile(distmu, npix),
            np.tile(distsigma, npix),
            np.tile(distnorm, npix),
        ),
        gps_time=1e9,
        objid="foobar",
        distmean=distmean,
        diststd=diststd,
    )
    return filename


@pytest.mark.parametrize("geo", [False, True])
@pytest.mark.parametrize("proj", ["mollweide", "aitoff", "globe", "zoom"])
def test_plot(tmpdir, skymap, geo, proj):
    """Test ligo-skymap-plot."""
    pngfilename = str(tmpdir / "skymap.png")
    args = [
        "ligo-skymap-plot",
        skymap,
        "-o",
        pngfilename,
        "--annotate",
        "--colorbar",
        "--radec",
        "320",
        "45",
        "--projection",
        proj,
    ]
    if geo:
        args.append("--geo")
    if proj == "globe":
        args.append("--projection-center")
        args.append("0d 0d" if geo else "0h 0d")
    elif proj == "zoom":
        args.append("--projection-center")
        args.append("0d 0d" if geo else "0h 0d")
        args.append("--zoom-radius")
        args.append("20deg")
    args.extend(["--contour", "90"])
    run_entry_point(*args)


def test_plot_volume(tmpdir, skymap):
    """Test ligo-skymap-plot-volume (at super-low resolution for speed)."""
    pngfilename = str(tmpdir / "skymap3d.png")
    args = [
        "ligo-skymap-plot-volume",
        skymap,
        "-o",
        pngfilename,
        "--dpi",
        "9",
        "--annotate",
        "--contour",
        "90",
        "--radecdist",
        "320",
        "45",
        "100",
    ]
    run_entry_point(*args)



import os
import pkgutil

__all__ = ()

# Import all symbols from all submodules of this module.
for _, module, _ in pkgutil.iter_modules([os.path.dirname(__file__)]):
    if module not in {"tests"}:
        exec(
            "from . import {0};"
            '__all__ += getattr({0}, "__all__", ());'
            "from .{0} import *".format(module)
        )
    del module

# Clean up
del os, pkgutil



#
# Copyright (C) 2013-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Functions that support the command line interface for plotting tools."""

import argparse
import os
import sys

import matplotlib

from . import HelpChoicesAction, type_with_sideeffect, version_string

# Set no-op Matplotlib backend to defer importing anything that requires a GUI
# until we have determined that it is necessary based on the command line
# arguments.
if "matplotlib.pyplot" in sys.modules:
    from matplotlib import pyplot as plt

    plt.switch_backend("Template")
else:
    matplotlib.use("Template", force=True)
    from matplotlib import pyplot as plt

__all__ = ("get_figure_parser",)


class MatplotlibFigureType(argparse.FileType):
    def __init__(self):
        super().__init__("wb")

    @staticmethod
    def __show():
        from matplotlib import pyplot as plt

        return plt.show()

    @staticmethod
    def get_savefig_metadata(format):
        program, _ = os.path.splitext(os.path.basename(sys.argv[0]))
        cmdline = " ".join([program] + sys.argv[1:])
        metadata = {"Title": cmdline}
        if format == "png":
            metadata["Software"] = version_string
        elif format in {"pdf", "ps", "eps"}:
            metadata["Creator"] = version_string
        return metadata

    def __save(self):
        from matplotlib import pyplot as plt

        _, ext = os.path.splitext(self.string)
        format = ext.lower().lstrip(".")
        metadata = self.get_savefig_metadata(format)
        return plt.savefig(self.string, metadata=metadata)

    def __call__(self, string):
        from matplotlib import pyplot as plt

        if string == "-":
            plt.switch_backend(matplotlib.rcParamsOrig["backend"])
            return self.__show
        else:
            with super().__call__(string):
                pass
            plt.switch_backend("agg")
            self.string = string
            return self.__save


@type_with_sideeffect(str)
def colormap(value):
    from matplotlib import rcParams

    rcParams["image.cmap"] = value


@type_with_sideeffect(float)
def figwidth(value):
    from matplotlib import rcParams

    rcParams["figure.figsize"][0] = float(value)


@type_with_sideeffect(float)
def figheight(value):
    from matplotlib import rcParams

    rcParams["figure.figsize"][1] = float(value)


@type_with_sideeffect(int)
def dpi(value):
    from matplotlib import rcParams

    rcParams["figure.dpi"] = rcParams["savefig.dpi"] = float(value)


@type_with_sideeffect(int)
def transparent(value):
    from matplotlib import rcParams

    rcParams["savefig.transparent"] = bool(value)


def get_figure_parser():
    parser = argparse.ArgumentParser(add_help=False)
    group = parser.add_argument_group(
        "figure options", "Options that affect figure output format"
    )
    group.add_argument(
        "-o",
        "--output",
        metavar="FILE.{pdf,png}",
        default="-",
        type=MatplotlibFigureType(),
        help="output file, or - to plot to screen",
    )
    group.add_argument(
        "--colormap",
        default="cylon",
        choices=plt.colormaps(),
        type=colormap,
        metavar="CMAP",
        help="matplotlib colormap",
    )
    group.add_argument(
        "--help-colormap", action=HelpChoicesAction, choices=plt.colormaps()
    )
    group.add_argument(
        "--figure-width",
        metavar="INCHES",
        type=figwidth,
        default="8",
        help="width of figure in inches",
    )
    group.add_argument(
        "--figure-height",
        metavar="INCHES",
        type=figheight,
        default="6",
        help="height of figure in inches",
    )
    group.add_argument(
        "--dpi",
        metavar="PIXELS",
        type=dpi,
        default=300,
        help="resolution of figure in dots per inch",
    )
    group.add_argument(
        "--transparent",
        const="1",
        default="0",
        nargs="?",
        type=transparent,
        help="Save image with transparent background",
    )
    return parser



#
# Copyright (C) 2019-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""Bullet charts for Bayes factors."""

import numpy as np
from matplotlib import pyplot as plt

__all__ = ("plot_bayes_factor",)


def plot_bayes_factor(
    logb,
    values=(1, 3, 5),
    labels=("", "strong", "very strong"),
    xlim=7,
    title=None,
    palette="RdYlBu",
    var_label="B",
):
    """Visualize a Bayes factor as a `bullet graph`_.

    Make a bar chart of a log Bayes factor as compared to a set of subjective
    threshold values. By default, use the thresholds from
    Kass & Raftery (1995).

    .. _`bullet graph`: https://en.wikipedia.org/wiki/Bullet_graph

    Parameters
    ----------
    logb : float
        The natural logarithm of the Bayes factor.
    values : list
        A list of floating point values for human-friendly confidence levels.
    labels : list
        A list of string labels for human-friendly confidence levels.
    xlim : float
        Limits of plot (`-xlim` to `+xlim`).
    title : str
        Title for plot.
    palette : str
        Color palette.
    var_label : str
        The variable symbol used in plotting

    Returns
    -------
    fig : Matplotlib figure
    ax : Matplotlib axes

    Examples
    --------

    .. plot::
       :include-source:

        from ligo.skymap.plot.bayes_factor import plot_bayes_factor
        plot_bayes_factor(6.3, title='BAYESTAR is awesome')

    """
    with plt.style.context("seaborn-v0_8-notebook"):
        fig, ax = plt.subplots(figsize=(6, 1.7), tight_layout=True)
        ax.set_xlim(-xlim, xlim)
        ax.set_ylim(-0.5, 0.5)
        ax.set_yticks([])
        ax.set_title(title)
        ax.set_ylabel(
            r"$\ln\,{}$".format(var_label),
            rotation=0,
            rotation_mode="anchor",
            ha="right",
            va="center",
        )

        # Add human-friendly labels
        ticks = (*(-x for x in reversed(values)), 0, *values)
        ticklabels = (
            *(f"{s}\nevidence\nagainst".strip() for s in reversed(labels)),
            "",
            *(f"{s}\nevidence\nfor".strip() for s in labels),
        )
        ax.set_xticks(ticks)
        ax.set_xticklabels(ticklabels)
        plt.setp(ax.get_xticklines(), visible=False)
        plt.setp(ax.get_xticklabels()[: len(ticks) // 2], ha="right")
        plt.setp(ax.get_xticklabels()[len(ticks) // 2 :], ha="left")

        # Plot colored bands for confidence thresholds
        fmt = plt.FuncFormatter(lambda x, _: f"{x:+g}".replace("+0", "0"))
        ax2 = ax.twiny()
        ax2.set_xlim(*ax.get_xlim())
        ax2.set_xticks(ticks)
        ax2.xaxis.set_major_formatter(fmt)
        levels = (-xlim, *ticks, xlim)
        colors = plt.get_cmap(palette)(np.arange(1, len(levels)) / len(levels))
        ax.barh(
            0,
            np.diff(levels),
            1,
            levels[:-1],
            linewidth=plt.rcParams["xtick.major.width"],
            color=colors,
            edgecolor="white",
        )

        # Plot bar for log Bayes factor value
        ax.barh(
            0,
            logb,
            0.5,
            color="black",
            linewidth=plt.rcParams["xtick.major.width"],
            edgecolor="white",
        )

        for ax_ in fig.axes:
            ax_.grid(False)
            for spine in ax_.spines.values():
                spine.set_visible(False)

    return fig, ax



#
# Copyright (C) 2013-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
r"""
Listen for new events from IGWN Alert and perform sky localization.

`bayestar-localize-lvalert` supports two modes of operation. You can
explicitly specify the GraceDb ID on the command line, as in::

    $ bayestar-localize-lvalert T90713

Or, `bayetar-localize-lvalert` can read GraceDB IDs from stdin (e.g., from the
terminal, or redirected from a fifo)::

    $ mkfifo /var/run/bayestar
    $ tail -F /var/run/bayestar | bayestar_localize_lvalert &
    $ echo T90713 > /var/run/bayestar
"""

from . import (
    ArgumentParser,
    EnableAction,
    get_mcmc_parser,
    get_posterior_parser,
    get_random_parser,
    get_waveform_parser,
    iterlines,
)


def parser():
    parser = ArgumentParser(
        parents=[
            get_waveform_parser(),
            get_posterior_parser(),
            get_mcmc_parser(),
            get_random_parser(),
        ]
    )
    parser.add_argument(
        "-d",
        "--disable-detector",
        metavar="X1",
        type=str,
        nargs="+",
        help="disable certain detectors",
    )
    parser.add_argument(
        "-N",
        "--dry-run",
        action="store_true",
        help="Dry run; do not update GraceDB entry",
    )
    parser.add_argument(
        "--no-tag", action="store_true", help="Do not set lvem tag for GraceDB entry"
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="FILE.fits[.gz]",
        default="bayestar.fits",
        help="Name for uploaded file",
    )
    parser.add_argument(
        "--enable-multiresolution",
        action=EnableAction,
        default=True,
        help="generate a multiresolution HEALPix map",
    )
    parser.add_argument(
        "graceid",
        metavar="G123456",
        nargs="*",
        help="Run on these GraceDB IDs. If no GraceDB IDs are listed on the "
        "command line, then read newline-separated GraceDB IDs from stdin.",
    )
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        import logging
        import os
        import re
        import sys
        import tempfile
        import urllib.parse

        import ligo.gracedb.logging
        import ligo.gracedb.rest

        from .. import omp
        from ..bayestar import localize, rasterize
        from ..io import events, fits
        from ..util.file import rename

        log = logging.getLogger("BAYESTAR")

        log.info("Using %d OpenMP thread(s)", omp.num_threads)

        # If no GraceDB IDs were specified on the command line, then read them
        # from stdin line-by-line.
        graceids = opts.graceid if opts.graceid else iterlines(sys.stdin)

        # Fire up a GraceDb client
        # FIXME: Mimic the behavior of the GraceDb command line client, where
        # the environment variable GRACEDB_SERVICE_URL overrides the default
        # service URL. It would be nice to get this behavior into the gracedb
        # package itself.
        gracedb = ligo.gracedb.rest.GraceDb(
            os.environ.get("GRACEDB_SERVICE_URL", ligo.gracedb.rest.DEFAULT_SERVICE_URL)
        )

        # Determine the base URL for event pages.
        scheme, netloc, *_ = urllib.parse.urlparse(gracedb._service_url)
        base_url = urllib.parse.urlunparse((scheme, netloc, "events", "", "", ""))

        if opts.chain_dump:
            chain_dump = re.sub(r".fits(.gz)?$", r".hdf5", opts.output)
        else:
            chain_dump = None

        tags = ("sky_loc",)
        if not opts.no_tag:
            tags += ("lvem",)

        event_source = events.gracedb.open(graceids, gracedb)

        if opts.disable_detector:
            event_source = events.detector_disabled.open(
                event_source, opts.disable_detector
            )

        for graceid in event_source.keys():
            try:
                event = event_source[graceid]
            except:  # noqa: E722
                log.exception("failed to read event %s from GraceDB", graceid)
                continue

            # Send log messages to GraceDb too
            if not opts.dry_run:
                handler = ligo.gracedb.logging.GraceDbLogHandler(gracedb, graceid)
                handler.setLevel(logging.INFO)
                logging.root.addHandler(handler)

            # A little bit of Cylon humor
            log.info("by your command...")

            try:
                # perform sky localization
                log.info("starting sky localization")
                sky_map = localize(
                    event,
                    opts.waveform,
                    opts.f_low,
                    opts.min_distance,
                    opts.max_distance,
                    opts.prior_distance_power,
                    opts.cosmology,
                    mcmc=opts.mcmc,
                    chain_dump=chain_dump,
                    enable_snr_series=opts.enable_snr_series,
                    f_high_truncate=opts.f_high_truncate,
                    rescale_loglikelihood=opts.rescale_loglikelihood,
                )
                if not opts.enable_multiresolution:
                    sky_map = rasterize(sky_map)
                sky_map.meta["objid"] = str(graceid)
                sky_map.meta["url"] = "{}/{}".format(base_url, graceid)
                log.info("sky localization complete")

                # upload FITS file
                with tempfile.TemporaryDirectory() as fitsdir:
                    fitspath = os.path.join(fitsdir, opts.output)
                    fits.write_sky_map(fitspath, sky_map, nest=True)
                    log.debug("wrote FITS file: %s", opts.output)
                    if opts.dry_run:
                        rename(fitspath, os.path.join(".", opts.output))
                    else:
                        gracedb.writeLog(
                            graceid,
                            "BAYESTAR rapid sky localization ready",
                            filename=fitspath,
                            tagname=tags,
                        )
                    log.debug("uploaded FITS file")
            except KeyboardInterrupt:
                # Produce log message and then exit if we receive SIGINT
                # (ctrl-C).
                log.exception("sky localization failed")
                raise
            except:  # noqa: E722
                # Produce log message for any otherwise uncaught exception.
                # Unless we are in dry-run mode, keep going.
                log.exception("sky localization failed")
                if opts.dry_run:
                    # Then re-raise the exception if we are in dry-run mode
                    raise

            if not opts.dry_run:
                # Remove old log handler
                logging.root.removeHandler(handler)
                del handler



# This file is adapted from the Astropy package template, which is licensed
# under a 3-clause BSD style license - see licenses/TEMPLATE_LICENSE.rst

# NOTE: The configuration for the package, including the name, version, and
# other information are set in the setup.cfg file.

import os

from setuptools import setup


def get_extensions():
    import numpy as np
    from extension_helpers import add_openmp_flags_if_available, pkg_config
    from setuptools import Extension

    pkg_config_packages = ["gsl"]

    orig_kwargs = {
        "sources": [
            "src/bayestar_distance.c",
            "src/bayestar_moc.c",
            "src/bayestar_sky_map.c",
            "src/core.c",
            "src/cubic_interp.c",
            "src/cubic_interp_test.c",
            "src/find_floor.c",
        ],
        "include_dirs": [np.get_include()],
        "define_macros": [
            ("GSL_RANGE_CHECK_OFF", None),
            ("HAVE_INLINE", None),
            ("Py_LIMITED_API", 0x030A0000),
            ("NPY_TARGET_VERSION", "NPY_1_23_API_VERSION"),
            ("NPY_NO_DEPRECATED_API", "NPY_2_0_API_VERSION"),
        ],
        "extra_compile_args": ["-std=gnu11", "-fvisibility=hidden"],
    }

    if os.environ.get("LIGO_SKYMAP_USE_SYSTEM_CHEALPIX"):
        pkg_config_packages.append("chealpix")
    else:
        orig_kwargs["include_dirs"].append("cextern/chealpix")
        orig_kwargs["sources"].append("cextern/chealpix/chealpix.c")

    if os.environ.get("LIGO_SKYMAP_USE_ITTNOTIFY"):
        pkg_config_packages.append("ittnotify")
        orig_kwargs["define_macros"].append(("WITH_ITTNOTIFY", 1))

    kwargs = pkg_config(pkg_config_packages, [])
    for key, orig_value in orig_kwargs.items():
        kwargs.setdefault(key, []).extend(orig_value)

    extension = Extension(
        name="ligo.skymap.core", language="c", py_limited_api=True, **kwargs
    )

    if not os.environ.get("LIGO_SKYMAP_DISABLE_OPENMP"):
        add_openmp_flags_if_available(extension)

    return [extension]


use_scm_version = {}

# If we are building under the GitLab CI pipeline and we are building on the
# default branch, then disable the local part of the version
# (+g<short commit hash>) so that we can upload nightly builds to PyPI.
if (
    os.environ.get("CI") == "true"
    and os.environ.get("CI_COMMIT_BRANCH") == os.environ["CI_DEFAULT_BRANCH"]
):
    use_scm_version["local_scheme"] = "no-local-version"

setup(use_scm_version=use_scm_version, ext_modules=get_extensions())



import numpy as np
import pytest
from astropy.table import Table

from . import run_entry_point


@pytest.mark.parametrize("ndatasets", [1, 2])
@pytest.mark.parametrize("groupby", [None, "far", "snr"])
def test_plot_stats(tmpdir, groupby, ndatasets):
    """Test ligo-skymap-plot-stats."""
    filenames = [str(tmpdir / "{}.out".format(i)) for i in range(ndatasets)]
    n = 250

    for filename in filenames:
        Table(
            {
                "searched_area": np.random.uniform(0, 1e4, size=n),
                "p_value": np.random.uniform(0, 1, size=n),
                "snr": np.random.uniform(10, 12, size=n),
                "far": 10 ** np.random.uniform(-12, -10, size=n),
            }
        ).write(filename, format="ascii.tab")

    args = ["ligo-skymap-plot-stats", "--format", "png", "-o", str(tmpdir)] + filenames
    if groupby is not None:
        args.extend(["--group-by", groupby])
    run_entry_point(*args)


def test_plot_stats_missing(tmpdir):
    """Test ligo-skymap-plot-stats, with group by value missing."""
    filename = str(tmpdir / "0.out")
    n = 250

    Table(
        {
            "p_value": np.random.uniform(0, 1, size=n),
        }
    ).write(filename, format="ascii.tab")

    args = ["ligo-skymap-plot-stats", "--group-by", "snr", "-o", str(tmpdir), filename]
    with pytest.raises(RuntimeError) as excinfo:
        run_entry_point(*args)
    assert 'The following files had no "snr" column' in str(excinfo.value)


def test_plot_stats_invalid(tmpdir):
    """Test ligo-skymap-plot-stats, with group by value invalid."""
    filename = str(tmpdir / "0.out")
    n = 250

    Table(
        {"p_value": np.random.uniform(0, 1, size=n), "snr": np.tile(np.nan, n)}
    ).write(filename, format="ascii.tab")

    args = ["ligo-skymap-plot-stats", "--group-by", "snr", "-o", str(tmpdir), filename]
    with pytest.raises(RuntimeError) as excinfo:
        run_entry_point(*args)
    assert 'The following files had invalid values in the "snr" column' in str(
        excinfo.value
    )



#
# Copyright (C) 2013-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Create summary plots for sky maps of found injections, optionally binned
cumulatively by false alarm rate or SNR.
"""

import os
from argparse import FileType

from . import ArgumentParser
from .matplotlib import MatplotlibFigureType


def parser():
    parser = ArgumentParser()
    parser.add_argument("--cumulative", action="store_true")
    parser.add_argument("--normed", action="store_true")
    parser.add_argument(
        "--group-by",
        choices=("far", "snr"),
        metavar="far|snr",
        help="Group plots by false alarm rate (FAR) or "
        + "signal to noise ratio (SNR)",
    )
    parser.add_argument(
        "--pp-confidence-interval",
        type=float,
        metavar="PCT",
        default=95,
        help="If all inputs files have the same number of "
        "samples, overlay binomial confidence bands for this percentage on "
        "the P--P plot",
    )
    parser.add_argument("--format", default="pdf", help="Matplotlib format")
    parser.add_argument(
        "input",
        type=FileType("rb"),
        nargs="+",
        help="Name of input file generated by ligo-skymap-stats",
    )
    parser.add_argument("--output", "-o", default=".", help="output directory")
    return parser


def main(args=None):
    with parser().parse_args(args) as opts:
        # Imports.
        import matplotlib
        from astropy.table import Table

        matplotlib.use("agg")
        import numpy as np
        from matplotlib import pyplot as plt
        from matplotlib import rcParams
        from tqdm import tqdm

        from .. import plot  # noqa: F401

        # Read in all of the datasets listed as positional command line
        # arguments.
        datasets = [Table.read(file, format="ascii") for file in opts.input]

        # Determine plot colors and labels
        filenames = [file.name for file in opts.input]
        labels = [os.path.splitext(os.path.basename(f))[0] for f in filenames]
        if rcParams["text.usetex"]:
            labels = [r"\verb/" + label + "/" for label in labels]
        rcParams["savefig.format"] = opts.format
        metadata = MatplotlibFigureType.get_savefig_metadata(opts.format)

        # Normalize column names
        for dataset in datasets:
            if "p_value" in dataset.colnames:
                dataset.rename_column("p_value", "searched_prob")

        if opts.group_by == "far":

            def key_func(table):
                return -np.log10(table["far"])

            def key_to_dir(key):
                return "far_1e{}".format(-key)

            def key_to_title(key):
                return r"$\mathrm{{FAR}} \leq 10^{{{}}}$ Hz".format(-key)

        elif opts.group_by == "snr":

            def key_func(table):
                return table["snr"]

            def key_to_dir(key):
                return "snr_{}".format(key)

            def key_to_title(key):
                return r"$\mathrm{{SNR}} \geq {}$".format(key)

        else:

            def key_func(table):
                return np.zeros(len(table))

            def key_to_dir(key):
                return "."

            def key_to_title(key):
                return "All events"

        if opts.group_by is not None:
            missing = [
                filename
                for filename, dataset in zip(filenames, datasets)
                if opts.group_by not in dataset.colnames
            ]
            if missing:
                raise RuntimeError(
                    'The following files had no "'
                    + opts.group_by
                    + '" column: '
                    + " ".join(missing)
                )

        for dataset in datasets:
            dataset["key"] = key_func(dataset)

        if opts.group_by is not None:
            invalid = [
                filename
                for filename, dataset in zip(filenames, datasets)
                if not np.all(np.isfinite(dataset["key"]))
            ]
            if invalid:
                raise RuntimeError(
                    'The following files had invalid values in the "'
                    + opts.group_by
                    + '" column: '
                    + " ".join(invalid)
                )

        keys = np.concatenate([dataset["key"] for dataset in datasets])

        histlabel = []
        if opts.cumulative:
            histlabel.append("cumulative")
        if opts.normed:
            histlabel.append("fraction")
        else:
            histlabel.append("number")
        histlabel.append("of injections")
        histlabel = " ".join(histlabel)

        pp_plot_settings = [
            ["", "searched posterior mass"],
            ["_dist", "distance CDF at true distance"],
            ["_vol", "searched volumetric probability"],
        ]
        hist_settings = [
            ["searched_area", "searched_area (deg$^2$)"],
            ["searched_vol", "searched volume (Mpc$^3$)"],
            ["offset", "angle from true location and mode of posterior (deg)"],
            ["runtime", "run time (s)"],
        ]

        keys = range(*np.floor([keys.min(), keys.max() + 1]).astype(int))
        total = len(keys) * (len(pp_plot_settings) + len(hist_settings))
        with tqdm(total=total) as progress:
            for key in keys:
                filtered = [d[d["key"] >= key] for d in datasets]
                title = key_to_title(key)
                nsamples = {len(d) for d in filtered}
                if len(nsamples) == 1:
                    (nsamples,) = nsamples
                    title += " ({} events)".format(nsamples)
                else:
                    nsamples = None

                subdir = os.path.join(opts.output, key_to_dir(key))
                os.makedirs(subdir, exist_ok=True)

                # Make several different kinds of P-P plots
                for suffix, xlabel in pp_plot_settings:
                    colname = "searched_prob" + suffix
                    fig = plt.figure(figsize=(6, 6))
                    ax = fig.add_subplot(111, projection="pp_plot")
                    fig.subplots_adjust(bottom=0.15)
                    ax.set_xlabel(xlabel)
                    ax.set_ylabel("cumulative fraction of injections")
                    ax.set_title(title)
                    for d, label in zip(filtered, labels):
                        ax.add_series(d.columns.get(colname, []), label=label)
                    ax.add_diagonal()
                    if nsamples:
                        ax.add_confidence_band(
                            nsamples, 0.01 * opts.pp_confidence_interval
                        )
                    ax.grid()
                    if len(filtered) > 1:
                        ax.legend(loc="lower right")
                    fig.savefig(os.path.join(subdir, colname), metadata=metadata)
                    plt.close()
                    progress.update()

                # Make several different kinds of histograms
                for colname, xlabel in hist_settings:
                    fig = plt.figure(figsize=(6, 4.5))
                    ax = fig.add_subplot(111)
                    fig.subplots_adjust(bottom=0.15)
                    ax.set_xscale("log")
                    ax.set_xlabel(xlabel)
                    ax.set_ylabel(histlabel)
                    ax.set_title(title)
                    values = np.concatenate(
                        [d.columns.get(colname, []) for d in filtered]
                    )
                    if len(values) > 0:
                        bins = np.geomspace(
                            np.min(values),
                            np.max(values),
                            1000 if opts.cumulative else 20,
                        )
                        for d, label in zip(filtered, labels):
                            ax.hist(
                                d.columns.get(colname, []),
                                cumulative=opts.cumulative,
                                density=opts.normed,
                                histtype="step",
                                bins=bins,
                                label=label,
                            )
                    ax.grid()
                    ax.legend(loc="upper left")
                    fig.savefig(
                        os.path.join(subdir, colname + "_hist"), metadata=metadata
                    )
                    plt.close()
                    progress.update()



#!/usr/bin/env python
#
# Copyright (C) 2020-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Combine two cobertura coverage files."""

import argparse

import lxml.etree

parser = argparse.ArgumentParser()
parser.add_argument("input1")
parser.add_argument("input2")
parser.add_argument("output")
args = parser.parse_args()

doc1 = lxml.etree.parse(args.input1)
doc2 = lxml.etree.parse(args.input2)
root1 = doc1.getroot()
root2 = doc2.getroot()
root1.attrib["lines-covered"] = str(
    int(root1.attrib["lines-covered"]) + int(root2.attrib["lines-covered"])
)
root1.attrib["lines-valid"] = str(
    int(root1.attrib["lines-valid"]) + int(root2.attrib["lines-valid"])
)
try:
    root1.attrib["line-rate"] = str(
        int(root1.attrib["lines-covered"]) / int(root1.attrib["lines-valid"])
    )
except ZeroDivisionError:
    root1.attrib["line-rate"] = "0"
root1.attrib["branches-covered"] = str(
    int(root1.attrib["branches-covered"]) + int(root2.attrib["branches-covered"])
)
root1.attrib["branches-valid"] = str(
    int(root1.attrib["branches-valid"]) + int(root2.attrib["branches-valid"])
)
try:
    root1.attrib["branch-rate"] = str(
        int(root1.attrib["branches-covered"]) / int(root1.attrib["branches-valid"])
    )
except ZeroDivisionError:
    root1.attrib["branch-rate"] = "0"
packages = root1.find("./packages")
packages.extend(root2.iterfind("./packages/package"))
doc1.write(args.output)



from subprocess import CalledProcessError

import numpy as np
import pytest
from astropy import units as u
from astropy.coordinates import (
    CartesianRepresentation,
    SkyCoord,
    SphericalRepresentation,
)
from astropy.table import Table
from astropy.utils.misc import NumpyRNGContext
from scipy import stats

from ...io.hdf5 import read_samples, write_samples
from . import run_entry_point


@pytest.fixture
def seed():
    with NumpyRNGContext(0):
        yield


@pytest.fixture
def samples(seed, tmpdir):
    mean = SkyCoord(
        ra=stats.uniform(0, 360).rvs() * u.hourangle,
        dec=np.arcsin(stats.uniform(-1, 1).rvs()) * u.radian,
        distance=stats.uniform(100, 200).rvs(),
    ).cartesian.xyz.value
    eigvals = stats.uniform(0, 1).rvs(3)
    eigvals *= len(eigvals) / eigvals.sum()
    cov = stats.random_correlation.rvs(eigvals) * 100
    pts = stats.multivariate_normal(mean, cov).rvs(200)
    pts = SkyCoord(pts, representation_type=CartesianRepresentation)
    pts.representation_type = SphericalRepresentation
    time = stats.uniform(-0.01, 0.01).rvs(200) + 1e9
    table = Table(
        {
            "ra": pts.ra.rad,
            "dec": pts.dec.rad,
            "distance": pts.distance.value,
            "time": time,
        }
    )
    filename = str(tmpdir / "samples.hdf5")
    write_samples(table, filename, path="/posterior_samples")
    return filename


@pytest.fixture
def samples_without_distance(samples, tmpdir):
    table = read_samples(samples, path="/posterior_samples")
    del table["dist"]
    filename = str(tmpdir / "samples_without_distance.hdf5")
    write_samples(table, filename, path="/posterior_samples")
    return filename


@pytest.mark.parametrize("enable_dpgmm", [False, True])
def test_from_samples(samples, tmpdir, enable_dpgmm):
    pytest.importorskip("figaro")
    """Test ligo-skymap-from-samples."""
    args = [
        "ligo-skymap-from-samples",
        "--seed",
        "150914",
        samples,
        "-o",
        str(tmpdir),
        "--instruments",
        "H1",
        "L1",
        "V1",
        "--objid",
        "S1234",
    ]

    if enable_dpgmm:
        args.append("--enable-dpgmm")
        args.append("--disable-distance-map")

    run_entry_point(*args)
    table = Table.read(str(tmpdir / "skymap.fits"), format="fits")
    assert table.meta["OBJECT"] == "S1234"
    assert table.meta["INSTRUME"] == "H1,L1,V1"


def test_from_samples_without_distance(samples_without_distance, capsys, tmpdir):
    """Test ligo-skymap-from-samples on an MCMC samples file that does not have
    a distance column.
    """
    with pytest.raises(CalledProcessError):
        run_entry_point(
            "ligo-skymap-from-samples", samples_without_distance, "-o", str(tmpdir)
        )
    out, err = capsys.readouterr()
    assert "does not have a distance column named 'dist' or 'distance'" in err



#
# Copyright (C) 2013-2025  Leo Singer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
"""Postprocessing utilities for HEALPix sky maps."""

import astropy_healpix as ah
import healpy as hp
import numpy as np
from astropy import units as u
from astropy.coordinates import (
    CartesianRepresentation,
    SkyCoord,
    UnitSphericalRepresentation,
)

__all__ = (
    "find_greedy_credible_levels",
    "interp_greedy_credible_levels",
    "smooth_ud_grade",
    "posterior_mean",
    "posterior_max",
)


def find_greedy_credible_levels(p, ranking=None):
    """Find the greedy credible levels of a (possibly multi-dimensional) array.

    Parameters
    ----------
    p : np.ndarray
        The input array, typically a HEALPix image.

    ranking : np.ndarray, optional
        The array to rank in order to determine the greedy order.
        The default is `p` itself.

    Returns
    -------
    cls : np.ndarray
        An array with the same shape as `p`, with values ranging from `0`
        to `p.sum()`, representing the greedy credible level to which each
        entry in the array belongs.

    """
    p = np.asarray(p)
    pflat = p.ravel()
    if ranking is None:
        ranking = pflat
    else:
        ranking = np.ravel(ranking)
    i = np.flipud(np.argsort(ranking))
    cs = np.cumsum(pflat[i])
    cls = np.empty_like(pflat)
    cls[i] = cs
    return cls.reshape(p.shape)


def smooth_ud_grade(m, nside, nest=False):
    """Resample a sky map to a new resolution using bilinear interpolation.

    Parameters
    ----------
    m : np.ndarray
        The input HEALPix array.

    nest : bool, default=False
        Indicates whether the input sky map is in nested rather than
        ring-indexed HEALPix coordinates (default: ring).

    Returns
    -------
    new_m : np.ndarray
        The resampled HEALPix array. The sum of `m` is approximately preserved.

    """
    npix = ah.nside_to_npix(nside)
    theta, phi = hp.pix2ang(nside, np.arange(npix), nest=nest)
    new_m = hp.get_interp_val(m, theta, phi, nest=nest)
    return new_m * len(m) / len(new_m)


def interp_greedy_credible_levels(x, xp, fp, right=None):
    """Perform linear interpolation suitable for finding credible levels.

    The linear interpolation is performed with the boundary condition that
    :math:`f(x) = 0`.

    Examples
    --------
    >>> xp = [1, 2, 3, 4, 5]
    >>> fp = [0.2, 0.4, 0.6, 0.8, 1.0]
    >>> interp_greedy_credible_levels([0, 0.5, 1.0, 1.5, 2.0], xp, fp)
    array([0. , 0.1, 0.2, 0.3, 0.4])
    """
    return np.interp(x, np.pad(xp, (1, 0)), np.pad(fp, (1, 0)), right=right)


def posterior_mean(prob, nest=False):
    npix = len(prob)
    nside = ah.npix_to_nside(npix)
    xyz = hp.pix2vec(nside, np.arange(npix), nest=nest)
    mean_xyz = np.average(xyz, axis=1, weights=prob)
    pos = SkyCoord(*mean_xyz, representation_type=CartesianRepresentation)
    pos.representation_type = UnitSphericalRepresentation
    return pos


def posterior_max(prob, nest=False):
    npix = len(prob)
    nside = ah.npix_to_nside(npix)
    i = np.argmax(prob)
    return SkyCoord(*hp.pix2ang(nside, i, nest=nest, lonlat=True), unit=u.deg)



import numpy as np
import pytest
from igwn_ligolw import lsctables as lsctables
from igwn_ligolw import utils as ligolw_utils
from lalinspiral.thinca import InspiralCoincDef

from ... import io
from ...io.events.ligolw import ContentHandler
from . import run_entry_point, run_lalsuite, run_ligolw


@pytest.fixture
def psd(tmpdir):
    filename = str(tmpdir / "psd.xml")
    run_entry_point(
        "bayestar-sample-model-psd",
        "-o",
        filename,
        "--H1=aLIGOZeroDetHighPower",
        "--L1=aLIGOZeroDetHighPower",
        "--V1=AdVDesignSensitivityP1200087",
    )
    return filename


@pytest.fixture
def inj(tmpdir):
    filename = str(tmpdir / "inj.xml")
    run_lalsuite(
        "lalapps_inspinj",
        "-o",
        filename,
        "--m-distr=fixMasses",
        "--fixed-mass1=1.4",
        "--fixed-mass2=1.4",
        "--gps-start-time=1000000000",
        "--gps-end-time=1000000001",
        "--time-step=1",
        "--l-distr=random",
        "--i-distr=uniform",
        "--d-distr=uniform",
        "--min-distance=1",
        "--max-distance=200e3",
        "--waveform=TaylorF2threePointFivePN",
        "--disable-spin",
        "--f-lower=10",
    )
    return filename


@pytest.fixture
def coinc(inj, psd, tmpdir):
    filename = str(tmpdir / "coinc.xml")
    run_entry_point(
        "bayestar-realize-coincs",
        "--measurement-error",
        "gaussian-noise",
        "--f-low",
        "25",
        "--reference-psd",
        psd,
        "-o",
        filename,
        inj,
        "--enable-snr-series",
        "--detector",
        "H1",
        "L1",
        "V1",
        "--seed",
        "42",
    )
    return filename


@pytest.fixture
def coinc_without_inj(coinc, tmpdir):
    """Produce a coinc.xml file with the found coincs stripped out."""
    filename = str(tmpdir / "coinc_without_inj.xml")
    xmldoc = ligolw_utils.load_filename(coinc, contenthandler=ContentHandler)

    # Prune coinc_def table
    coinc_def_table = lsctables.CoincDefTable.get_table(xmldoc)
    included = [
        row
        for row in coinc_def_table
        if row.search_coinc_type == InspiralCoincDef.search_coinc_type
        and row.search == InspiralCoincDef.search
    ]
    included_coinc_def_ids = {row.coinc_def_id for row in included}
    coinc_def_table[:] = included

    # Prune coinc table
    coinc_table = lsctables.CoincTable.get_table(xmldoc)
    included = [
        row for row in coinc_table if row.coinc_def_id in included_coinc_def_ids
    ]
    included_coinc_ids = {row.coinc_event_id for row in included}
    coinc_table[:] = included

    # Prune coinc_map table
    coinc_map_table = lsctables.CoincMapTable.get_table(xmldoc)
    coinc_map_table[:] = [
        row for row in coinc_map_table if row.coinc_event_id in included_coinc_ids
    ]

    ligolw_utils.write_filename(xmldoc, filename)
    return filename


@pytest.fixture
def coinc_sqlite(coinc, tmpdir):
    filename = str(tmpdir / "coinc.sqlite")
    run_ligolw("igwn_ligolw_sqlite", coinc, "-p", "-r", "-d", filename)
    return filename


@pytest.fixture
def localize_coincs(coinc, psd, tmpdir):
    run_entry_point("bayestar-localize-coincs", coinc, "-o", str(tmpdir))
    return str(tmpdir / "0.fits")


# Note: any test that uses this fixture should be marked with
# @pytest.mark.internet_off to make sure that it does not actually
# contact GraceDb.
@pytest.fixture
def localize_lvalert(coinc_without_inj, psd, tmpdir, monkeypatch):
    class MockGraceDb:
        def __init__(self, service_url, *args, **kwargs):
            self._service_url = service_url

        def files(self, graceid, filename):
            assert graceid == "G1234"
            mock_filename = {"coinc.xml": coinc_without_inj, "psd.xml.gz": psd}[
                filename
            ]
            return open(mock_filename, "rb")

        def writeLog(self, *args, **kwargs):  # noqa: N802
            pass

    monkeypatch.setattr("ligo.gracedb.rest.GraceDb", MockGraceDb)

    filename = str(tmpdir / "bayestar.fits")
    run_entry_point("bayestar-localize-lvalert", "G1234", "-N", "-o", filename)
    return filename


@pytest.mark.internet_off
def test_bayestar(localize_coincs, localize_lvalert, coinc_sqlite, tmpdir):
    """Test bayestar-realize-coincs, bayestar-localize-coincs,
    bayestar-localize-lvalert, and ligo-skymap-stats.
    """
    # Check that bayestar-localize-coincs and bayestar-localize-lvalert
    # produce the same output.
    skymap1, meta1 = io.read_sky_map(localize_coincs, distances=True)
    skymap2, meta2 = io.read_sky_map(localize_lvalert, distances=True)
    for col1, col2 in zip(skymap1, skymap2):
        np.testing.assert_allclose(col1, col2)
    for key in "gps_time origin vcs_version".split():
        assert meta1[key] == meta2[key]
    for key in "distmean diststd log_bci log_bsn".split():
        np.testing.assert_allclose(meta1[key], meta2[key])

    # Test ligo-skymap-stats.
    out1 = str(tmpdir / "stats1.out")
    out2 = str(tmpdir / "stats2.out")
    args = ("ligo-skymap-stats", "--modes", "-p", "90", "-a", "100", "-o")
    run_entry_point(*args, out1, localize_coincs, "-d", coinc_sqlite)
    run_entry_point(*args, out2, localize_lvalert, "-j", "2")
